"""Group all anagrams together in a list.

Anagram is a word formed by rearranging the letters of a different word, using all the original letters exactly once.

A popular Leetcode problem:
https://leetcode.com/problems/group-anagrams/description/
"""

import json
from collections import defaultdict
from dataclasses import dataclass
from random import Random
from typing import Any, Optional

from ..coaching import BaseCurriculum, RangeAttributeDefinition
from ..data import get_data_file_path
from ..factory import ProceduralDataset, register_dataset

QUESTION_TEMPLATE = """An anagram is a word formed by rearranging the letters of a different word, using all the original letters exactly once.

Your job is to group the anagrams together. You can return the answer in any order.

The output is a list of lists of strings, where each outer list contains a group of anagrams, e.g. [["eat", "tea"], ["tan", "nat"]].

Group the following list of words into anagrams:
{words}
"""

DATASET_NAME = "group_anagrams"


@dataclass
class GroupAnagramsConfig:
    """Configuration for Group Anagrams dataset generation"""

    min_anagram_groups: int = 2  # Minimum number of anagram groups present in the input
    max_anagram_groups: int = 10  # Maximum number of anagram groups present in the input
    min_words_per_group: int = 2  # Minimum number of words in a single anagram group
    max_words_per_group: int = 5  # Maximum number of words in a single anagram group

    size: int = 500  # Virtual dataset size
    seed: Optional[int] = None

    def validate(self):
        """Validate configuration parameters"""
        assert 2 <= self.min_anagram_groups <= self.max_anagram_groups, "Invalid number of anagram groups"
        assert 2 <= self.min_words_per_group <= self.max_words_per_group, "Invalid number of words per group"


class GroupAnagramsDataset(ProceduralDataset):
    """Generates Group Anagrams exercises with configurable difficulty"""

    def __init__(self, config: GroupAnagramsConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)
        with get_data_file_path("anagrams.jsonl").open() as f:
            self.anagrams = [json.loads(line)["words"] for line in f]

    def _get_anagram_words(self, rng: Random, num_groups: int) -> list[str]:
        """Generate a list of words with anagrams"""
        words = []
        for sample in rng.sample(self.anagrams, num_groups):
            num_words = min(len(sample), rng.randint(self.config.min_words_per_group, self.config.max_words_per_group))
            anagrams = rng.sample(sample, num_words)
            words.extend(anagrams)
        return words

    def _sort_nested_list(self, lst: list[list[str]]) -> list[list[str]]:
        """Sort a nested list of strings"""
        return sorted([sorted(sublist) for sublist in lst], key=lambda x: x[0] if x else "")

    def _group_anagrams(self, words: list[str]) -> list[list[str]]:
        """Group anagrams together"""

        def _codify(word):
            code = [0] * 26
            for c in word:
                code[ord(c) - ord("a")] += 1
            return tuple(code)

        res = defaultdict(list)
        for word in words:
            code = _codify(word)
            res[code].append(word)

        anagrams = list(res.values())
        return self._sort_nested_list(anagrams)

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        """Score a single Group Anagrams question"""
        reward = 0.0
        if answer is not None:
            try:
                answer = json.loads(answer)
                oracle = entry["metadata"]["solution"]
                answer_str = json.dumps(self._sort_nested_list(answer))
                oracle_str = json.dumps(self._sort_nested_list(oracle))
                if answer_str == oracle_str:
                    reward = 1.0
                else:
                    reward = 0.01  # json parsable
            except Exception:
                reward = 0.0
        return reward

    def __getitem__(self, idx: int) -> dict:
        """Generate a single Group Anagrams question"""
        rng = Random(self.seed + idx)

        anagram_groups = min(
            len(self.anagrams), rng.randint(self.config.min_anagram_groups, self.config.max_anagram_groups)
        )
        words = self._get_anagram_words(rng, num_groups=anagram_groups)
        answer = self._group_anagrams(words)
        answer_str = json.dumps(answer)

        return {
            "question": QUESTION_TEMPLATE.format(words=json.dumps(words)),
            "answer": answer_str,
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "words": words,
                "solution": answer,
                "anagram_groups": anagram_groups,
                "difficulty": {
                    "anagram_groups": (self.config.min_anagram_groups, self.config.max_anagram_groups),
                    "words_per_group": (self.config.min_words_per_group, self.config.max_words_per_group),
                },
            },
        }


class GroupAnagramsCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(GroupAnagramsCurriculum.__name__, GroupAnagramsConfig)

        # Define attributes
        self._define_attributes(
            RangeAttributeDefinition(
                name="anagram_groups",
                levels=[5, 10, 50, 100],
                description="Number of anagram groups in the input",
                lower_field_name="min_anagram_groups",
                upper_field_name="max_anagram_groups",
                ensure_interval=True,
            ),
            RangeAttributeDefinition(
                name="words_per_group",
                levels=[2, 5, 10],
                description="Number of words in a single anagram group",
                lower_field_name="min_words_per_group",
                upper_field_name="max_words_per_group",
            ),
        )


register_dataset(DATASET_NAME, GroupAnagramsDataset, GroupAnagramsConfig, GroupAnagramsCurriculum)



from dataclasses import dataclass, field
from random import Random
from string import Template
from typing import Optional

from ..coaching import BaseCurriculum, ScalarAttributeDefinition
from ..factory import ProceduralDataset, register_dataset
from ..utils import StrEnum

DATASET_NAME = "aiw"


class TaskType(StrEnum):
    """Defines the type of task for the Alice in Wonderland dataset."""

    SIBLINGS = "siblings"
    FRIENDS = "friends"
    COLLEAGUES = "colleagues"  # Added colleagues task


@dataclass
class AliceInWonderlandConfig:
    """Configuration options for the Alice in Wonderland dataset.

    Attributes:
        male_names (list[str]): List of male names to use in questions.
        female_names (list[str]): List of female names to use in questions. Must include 'Alice'.
        task_types (list[TaskType]): List of task types to include in dataset.
        seed (Optional[int]): Seed for random number generation.
        size (int): Number of samples in the dataset.
        max_entities (int): Max number of siblings/friends/colleagues in questions.
    """

    male_names: list[str] = field(
        default_factory=lambda: [
            "James",
            "John",
            "Robert",
            "Michael",
            "William",
            "David",
            "Richard",
            "Joseph",
            "Thomas",
            "Charles",
            "Bob",
        ]
    )
    female_names: list[str] = field(
        default_factory=lambda: [
            "Mary",
            "Patricia",
            "Jennifer",
            "Linda",
            "Elizabeth",
            "Barbara",
            "Susan",
            "Jessica",
            "Sarah",
            "Margaret",
            "Alice",
        ]
    )
    task_types: list[TaskType] = field(
        default_factory=lambda: [TaskType.SIBLINGS, TaskType.FRIENDS, TaskType.COLLEAGUES]  # Added Colleagues
    )
    task_type_weights: list[float] = field(default_factory=lambda: [1 / 3, 1 / 3, 1 / 3])
    seed: Optional[int] = None
    size: int = 10
    max_entities: int = 6  # Added max_entities

    def validate(self) -> None:
        """Validates the configuration parameters."""
        assert len(self.male_names) > 0, "must provide male names"
        assert len(self.female_names) > 0, "must provide female names"
        assert "Alice" in self.female_names, "'Alice' must be in female names"
        assert len(self.task_types) > 0, "must provide at least one task type"
        assert self.max_entities > 0, "max_entities must be positive"


class AliceInWonderlandDataset(ProceduralDataset):
    """
    A procedural dataset inspired by the "Alice in Wonderland" paper.

    The dataset is inspired by the following paper:
       @inproceedings{nezhurina2024alice,
       title={Alice in Wonderland: Simple Tasks Reveal Severe Generalization and
              Basic Reasoning Deficits in State-Of-the-Art Large Language Models},
       author={Marianna Nezhurina and Lucia Cipolina-Kun and Mehdi Cherti and
              Jenia Jitsev},
       booktitle={NeurIPS 2024 Workshop on Scientific Methods for Understanding
                  Deep Learning},
       year={2024},
       url={https://openreview.net/forum?id=Mkl7dzjYiW}
       }

    """

    def __init__(self, config: AliceInWonderlandConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)
        self.templates = {
            TaskType.SIBLINGS: [
                Template(
                    "$female_name has $num_brothers brothers and she also has "
                    "$num_sisters sisters. How many sisters does "
                    "$female_name's brother have?"
                ),
                Template(
                    "$female_name has $num_sisters sisters and she also has "
                    "$num_brothers brothers. How many sisters does "
                    "$male_name's brother have?"
                ),
            ],
            TaskType.FRIENDS: [
                Template(
                    "$female_name has $num_male male friends and she also has "
                    "$num_female female friends. They all are friends with each "
                    "other and have no other friends aside. How many female "
                    "friends does $male_name, a male friend of $female_name, "
                    "have?"
                )
            ],
            TaskType.COLLEAGUES: [  # New colleagues templates
                Template(
                    "$female_name has $num_male_colleagues_alice_circle male colleagues and she also has "
                    "$num_female_colleagues_alice_circle female colleagues. These are all colleagues that $female_name has. "
                    "All these mentioned persons around $female_name are colleagues of each other. "
                    "$male_name has $num_male_colleagues_bob_circle male colleagues "
                    "and $num_female_colleagues_bob_circle female colleagues in total. "
                    "All these mentioned persons around $male_name are colleagues of each other. "
                    "The people in the circle around $male_name do not have "
                    "other colleagues aside - with the only exception of Matilda. "
                    "She is colleague of $male_name and she is also colleague of $female_name, "
                    "being part of $female_name's circle. How many female colleagues does Matilda have?"
                ),
            ],
        }

    def _get_aiw(self, rng: Random, idx: int) -> dict:
        """Generates a single Alice in Wonderland question.

        Args:
            rng (Random): Random number generator.

        Returns:
            dict: A dictionary containing the generated question, the right answer
                and a description of the example.
        """

        task_type = rng.choices(self.config.task_types, weights=self.config.task_type_weights, k=1)[0]
        female_name = rng.choice(self.config.female_names)
        male_name = rng.choice(self.config.male_names)

        if task_type == TaskType.SIBLINGS:
            num_brothers = rng.randint(1, self.config.max_entities)
            num_sisters = rng.randint(1, self.config.max_entities)

            answer = num_sisters + 1
            template = rng.choice(self.templates[TaskType.SIBLINGS])
            question = template.substitute(
                female_name=female_name,
                male_name=male_name,
                num_brothers=num_brothers,
                num_sisters=num_sisters,
            )
        elif task_type == TaskType.FRIENDS:
            num_male = rng.randint(1, self.config.max_entities)
            num_female = rng.randint(1, self.config.max_entities)

            answer = num_female + 1
            template = rng.choice(self.templates[TaskType.FRIENDS])
            question = template.substitute(
                female_name=female_name,
                male_name=male_name,
                num_male=num_male,
                num_female=num_female,
            )
        elif task_type == TaskType.COLLEAGUES:
            num_male_colleagues_alice_circle = rng.randint(1, self.config.max_entities)
            num_female_colleagues_alice_circle = rng.randint(1, self.config.max_entities)
            num_male_colleagues_bob_circle = rng.randint(1, self.config.max_entities)
            num_female_colleagues_bob_circle = rng.randint(1, self.config.max_entities)

            answer = num_female_colleagues_alice_circle + 1
            template = rng.choice(self.templates[TaskType.COLLEAGUES])
            question = template.substitute(
                female_name=female_name,
                male_name=male_name,
                num_male_colleagues_alice_circle=num_male_colleagues_alice_circle,
                num_female_colleagues_alice_circle=num_female_colleagues_alice_circle,
                num_male_colleagues_bob_circle=num_male_colleagues_bob_circle,
                num_female_colleagues_bob_circle=num_female_colleagues_bob_circle,
            )

        return {
            "question": question,
            "answer": str(answer),
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "task_type": task_type.value,
                "difficulty": {
                    "task_type_weights": self.config.task_type_weights,
                    "num_entities": self.config.max_entities,
                },
            },
        }

    def __getitem__(self, idx: int) -> dict:
        rng = Random(self.seed + idx)
        return self._get_aiw(rng, idx)


class AliceInWonderlandCurriculum(BaseCurriculum):
    """Curriculum for the Alice in Wonderland dataset."""

    def __init__(self):
        super().__init__(AliceInWonderlandCurriculum.__name__, AliceInWonderlandConfig)
        self._define_attributes(
            ScalarAttributeDefinition(
                name="task_type_weights",
                field_name="task_type_weights",
                description="The weight of the task type",
                levels=[
                    [1.0, 0.0, 0.0],
                    [0.9, 0.05, 0.05],
                    [0.7, 0.15, 0.15],
                    [0.6, 0.2, 0.2],
                    [0.5, 0.25, 0.25],
                    [0.4, 0.3, 0.3],
                    [0.3, 0.35, 0.35],
                    [0.2, 0.4, 0.4],
                    [0.1, 0.45, 0.45],
                ],
            ),
            ScalarAttributeDefinition(
                name="num_entities",
                field_name="max_entities",
                description="The number of entities in the question",
                levels=list(range(4, 18, 2)),
            ),
        )


register_dataset(DATASET_NAME, AliceInWonderlandDataset, AliceInWonderlandConfig, AliceInWonderlandCurriculum)



"""Word sorting task generator"""

import re
from dataclasses import dataclass
from random import Random
from typing import Any, Optional

from ..coaching import BaseCurriculum, RangeAttributeDefinition
from ..data import read_data_file
from ..factory import ProceduralDataset, register_dataset
from ..utils import StrEnum


class TextTransformation(StrEnum):
    """Text transformation options"""

    LOWERCASE = "lowercase"
    UPPERCASE = "uppercase"
    ORIGINAL = "original"
    RANDOMCASE = "randomcase"


QUESTION_TEMPLATE = """Your task is to sort words in ascending or descending order using ASCII/Unicode ordering.

Your output should be a comma-separated list of words, e.g. word_1, word_2, word_3

Now, sort these words in {direction} order (using ASCII/Unicode ordering) and return them as a comma-separated list: {words}
"""

DATASET_NAME = "word_sorting"


@dataclass
class WordSortingConfig:
    """Configuration for word sorting task generation"""

    min_words: int = 3  # Minimum words to sort
    max_words: int = 10  # Maximum words to sort
    min_word_length: int = 3  # Minimum word length
    max_word_length: int = 12  # Maximum word length
    transformation: TextTransformation = TextTransformation.ORIGINAL
    seed: Optional[int] = None
    size: int = 500  # Virtual dataset size

    def validate(self) -> None:
        """Validate configuration parameters"""
        assert self.min_words > 0, "min_words must be positive"
        assert self.min_words <= self.max_words, "max_words must be >= min_words"
        assert self.min_word_length > 0, "min_word_length must be positive"
        assert self.min_word_length <= self.max_word_length, "max_word_length must be >= min_word_length"
        assert isinstance(self.transformation, TextTransformation), "transformation must be a TextTransformation"


class WordSortingDataset(ProceduralDataset):
    """Generates word sorting tasks"""

    def __init__(self, config: WordSortingConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)

        # Load and preprocess text
        text = read_data_file("in_the_year_2889.txt")
        # Extract unique words within length constraints
        self.words = sorted(
            set(
                word
                for word in re.findall(r"\b\w+\b", text)
                if self.config.min_word_length <= len(word) <= self.config.max_word_length
            )
        )

    def _transform_word(self, word: str, rng: Random) -> str:
        """Apply configured transformation to word"""
        if self.config.transformation == TextTransformation.LOWERCASE:
            return word.lower()
        elif self.config.transformation == TextTransformation.UPPERCASE:
            return word.upper()
        elif self.config.transformation == TextTransformation.RANDOMCASE:
            return "".join(c.upper() if rng.choice([True, False]) else c.lower() for c in word)
        return word  # ORIGINAL case

    def _generate_words(self, rng: Random) -> tuple[list[str], list[str]]:
        """Generate list of words and their transformed versions"""
        count = rng.randint(self.config.min_words, self.config.max_words)

        # Select random words
        selected_words = rng.sample(self.words, count)
        # Apply transformation
        transformed_words = [self._transform_word(word, rng) for word in selected_words]

        return selected_words, transformed_words

    def __getitem__(self, idx: int) -> dict:
        """Generate a single sorting task"""
        rng = Random(self.seed + idx)

        original_words, transformed_words = self._generate_words(rng)

        # Generate both ascending and descending answers
        asc_words = sorted(transformed_words)
        desc_words = sorted(transformed_words, reverse=True)

        # Randomly choose ascending or descending
        is_ascending = rng.choice([True, False])
        direction = "ascending" if is_ascending else "descending"
        answer = asc_words if is_ascending else desc_words

        return {
            "question": QUESTION_TEMPLATE.format(direction=direction, words=", ".join(transformed_words)),
            "answer": ", ".join(answer),
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "original_words": original_words,
                "sorted_words": answer,
                "transformed_words": transformed_words,
                "direction": direction,
                "num_words": len(original_words),
                "word_length": max(len(word) for word in original_words),
                "difficulty": {
                    "num_words": (self.config.min_words, self.config.max_words),
                    "word_length": (self.config.min_word_length, self.config.max_word_length),
                },
            },
        }

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        oracle_answer = entry["metadata"]["sorted_words"]

        if not answer:
            return 0.0

        parsed_answer = [word.strip() for word in re.split(r",\s*", answer)]

        if parsed_answer == oracle_answer:
            return 1.0

        correct_positions = sum(
            1 for i, word in enumerate(parsed_answer) if i < len(oracle_answer) and word == oracle_answer[i]
        )

        partial_score = correct_positions / len(oracle_answer)

        if sorted(parsed_answer) == sorted(oracle_answer):
            partial_score = max(partial_score, 0.2)

        return partial_score


class WordSortingCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(WordSortingCurriculum.__name__, WordSortingConfig)

        self._define_attributes(
            RangeAttributeDefinition(
                name="num_words",
                levels=[5, 10, 25, 50, 100],
                description="Number of words to sort",
                lower_field_name="min_words",
                upper_field_name="max_words",
                ensure_interval=True,
            ),
            RangeAttributeDefinition(
                name="word_length",
                levels=[3, 5, 10, 15],
                description="Length of words to sort",
                lower_field_name="min_word_length",
                upper_field_name="max_word_length",
                ensure_interval=True,
            ),
        )


register_dataset(DATASET_NAME, WordSortingDataset, WordSortingConfig, WordSortingCurriculum)



NAMES = [
    "Aaran",
    "Aaren",
    "Aarez",
    "Aarman",
    "Aaron",
    "Aaron-James",
    "Aarron",
    "Aaryan",
    "Aaryn",
    "Aayan",
    "Aazaan",
    "Abaan",
    "Abbas",
    "Abdallah",
    "Abdalroof",
    "Abdihakim",
    "Abdirahman",
    "Abdisalam",
    "Abdul",
    "Abdul-Aziz",
    "Abdulbasir",
    "Abdulkadir",
    "Abdulkarem",
    "Abdulkhader",
    "Abdullah",
    "Abdul-Majeed",
    "Abdulmalik",
    "Abdul-Rehman",
    "Abdur",
    "Abdurraheem",
    "Abdur-Rahman",
    "Abdur-Rehmaan",
    "Abel",
    "Abhinav",
    "Abhisumant",
    "Abid",
    "Abir",
    "Abraham",
    "Abu",
    "Abubakar",
    "Ace",
    "Adain",
    "Adam",
    "Adam-James",
    "Addison",
    "Addisson",
    "Adegbola",
    "Adegbolahan",
    "Aden",
    "Adenn",
    "Adie",
    "Adil",
    "Aditya",
    "Adnan",
    "Adrian",
    "Adrien",
    "Aedan",
    "Aedin",
    "Aedyn",
    "Aeron",
    "Afonso",
    "Ahmad",
    "Ahmed",
    "Ahmed-Aziz",
    "Ahoua",
    "Ahtasham",
    "Aiadan",
    "Aidan",
    "Aiden",
    "Aiden-Jack",
    "Aiden-Vee",
    "Aidian",
    "Aidy",
    "Ailin",
    "Aiman",
    "Ainsley",
    "Ainslie",
    "Airen",
    "Airidas",
    "Airlie",
    "AJ",
    "Ajay",
    "A-Jay",
    "Ajayraj",
    "Akan",
    "Akram",
    "Al",
    "Ala",
    "Alan",
    "Alanas",
    "Alasdair",
    "Alastair",
    "Alber",
    "Albert",
    "Albie",
    "Aldred",
    "Alec",
    "Aled",
    "Aleem",
    "Aleksandar",
    "Aleksander",
    "Aleksandr",
    "Aleksandrs",
    "Alekzander",
    "Alessandro",
    "Alessio",
    "Alex",
    "Alexander",
    "Alexei",
    "Alexx",
    "Alexzander",
    "Alf",
    "Alfee",
    "Alfie",
    "Alfred",
    "Alfy",
    "Alhaji",
    "Al-Hassan",
    "Ali",
    "Aliekber",
    "Alieu",
    "Alihaider",
    "Alisdair",
    "Alishan",
    "Alistair",
    "Alistar",
    "Alister",
    "Aliyaan",
    "Allan",
    "Allan-Laiton",
    "Allen",
    "Allesandro",
    "Allister",
    "Ally",
    "Alphonse",
    "Altyiab",
    "Alum",
    "Alvern",
    "Alvin",
    "Alyas",
    "Amaan",
    "Aman",
    "Amani",
    "Ambanimoh",
    "Ameer",
    "Amgad",
    "Ami",
    "Amin",
    "Amir",
    "Ammaar",
    "Ammar",
    "Ammer",
    "Amolpreet",
    "Amos",
    "Amrinder",
    "Amrit",
    "Amro",
    "Anay",
    "Andrea",
    "Andreas",
    "Andrei",
    "Andrejs",
    "Andrew",
    "Andy",
    "Anees",
    "Anesu",
    "Angel",
    "Angelo",
    "Angus",
    "Anir",
    "Anis",
    "Anish",
    "Anmolpreet",
    "Annan",
    "Anndra",
    "Anselm",
    "Anthony",
    "Anthony-John",
    "Antoine",
    "Anton",
    "Antoni",
    "Antonio",
    "Antony",
    "Antonyo",
    "Anubhav",
    "Aodhan",
    "Aon",
    "Aonghus",
    "Apisai",
    "Arafat",
    "Aran",
    "Arandeep",
    "Arann",
    "Aray",
    "Arayan",
    "Archibald",
    "Archie",
    "Arda",
    "Ardal",
    "Ardeshir",
    "Areeb",
    "Areez",
    "Aref",
    "Arfin",
    "Argyle",
    "Argyll",
    "Ari",
    "Aria",
    "Arian",
    "Arihant",
    "Aristomenis",
    "Aristotelis",
    "Arjuna",
    "Arlo",
    "Armaan",
    "Arman",
    "Armen",
    "Arnab",
    "Arnav",
    "Arnold",
    "Aron",
    "Aronas",
    "Arran",
    "Arrham",
    "Arron",
    "Arryn",
    "Arsalan",
    "Artem",
    "Arthur",
    "Artur",
    "Arturo",
    "Arun",
    "Arunas",
    "Arved",
    "Arya",
    "Aryan",
    "Aryankhan",
    "Aryian",
    "Aryn",
    "Asa",
    "Asfhan",
    "Ash",
    "Ashlee-jay",
    "Ashley",
    "Ashton",
    "Ashton-Lloyd",
    "Ashtyn",
    "Ashwin",
    "Asif",
    "Asim",
    "Aslam",
    "Asrar",
    "Ata",
    "Atal",
    "Atapattu",
    "Ateeq",
    "Athol",
    "Athon",
    "Athos-Carlos",
    "Atli",
    "Atom",
    "Attila",
    "Aulay",
    "Aun",
    "Austen",
    "Austin",
    "Avani",
    "Averon",
    "Avi",
    "Avinash",
    "Avraham",
    "Awais",
    "Awwal",
    "Axel",
    "Ayaan",
    "Ayan",
    "Aydan",
    "Ayden",
    "Aydin",
    "Aydon",
    "Ayman",
    "Ayomide",
    "Ayren",
    "Ayrton",
    "Aytug",
    "Ayub",
    "Ayyub",
    "Azaan",
    "Azedine",
    "Azeem",
    "Azim",
    "Aziz",
    "Azlan",
    "Azzam",
    "Azzedine",
    "Babatunmise",
    "Babur",
    "Bader",
    "Badr",
    "Badsha",
    "Bailee",
    "Bailey",
    "Bailie",
    "Bailley",
    "Baillie",
    "Baley",
    "Balian",
    "Banan",
    "Barath",
    "Barkley",
    "Barney",
    "Baron",
    "Barrie",
    "Barry",
    "Bartlomiej",
    "Bartosz",
    "Basher",
    "Basile",
    "Baxter",
    "Baye",
    "Bayley",
    "Beau",
    "Beinn",
    "Bekim",
    "Believe",
    "Ben",
    "Bendeguz",
    "Benedict",
    "Benjamin",
    "Benjamyn",
    "Benji",
    "Benn",
    "Bennett",
    "Benny",
    "Benoit",
    "Bentley",
    "Berkay",
    "Bernard",
    "Bertie",
    "Bevin",
    "Bezalel",
    "Bhaaldeen",
    "Bharath",
    "Bilal",
    "Bill",
    "Billy",
    "Binod",
    "Bjorn",
    "Blaike",
    "Blaine",
    "Blair",
    "Blaire",
    "Blake",
    "Blazej",
    "Blazey",
    "Blessing",
    "Blue",
    "Blyth",
    "Bo",
    "Boab",
    "Bob",
    "Bobby",
    "Bobby-Lee",
    "Bodhan",
    "Boedyn",
    "Bogdan",
    "Bohbi",
    "Bony",
    "Bowen",
    "Bowie",
    "Boyd",
    "Bracken",
    "Brad",
    "Bradan",
    "Braden",
    "Bradley",
    "Bradlie",
    "Bradly",
    "Brady",
    "Bradyn",
    "Braeden",
    "Braiden",
    "Brajan",
    "Brandan",
    "Branden",
    "Brandon",
    "Brandonlee",
    "Brandon-Lee",
    "Brandyn",
    "Brannan",
    "Brayden",
    "Braydon",
    "Braydyn",
    "Breandan",
    "Brehme",
    "Brendan",
    "Brendon",
    "Brendyn",
    "Breogan",
    "Bret",
    "Brett",
    "Briaddon",
    "Brian",
    "Brodi",
    "Brodie",
    "Brody",
    "Brogan",
    "Broghan",
    "Brooke",
    "Brooklin",
    "Brooklyn",
    "Bruce",
    "Bruin",
    "Bruno",
    "Brunon",
    "Bryan",
    "Bryce",
    "Bryden",
    "Brydon",
    "Brydon-Craig",
    "Bryn",
    "Brynmor",
    "Bryson",
    "Buddy",
    "Bully",
    "Burak",
    "Burhan",
    "Butali",
    "Butchi",
    "Byron",
    "Cabhan",
    "Cadan",
    "Cade",
    "Caden",
    "Cadon",
    "Cadyn",
    "Caedan",
    "Caedyn",
    "Cael",
    "Caelan",
    "Caelen",
    "Caethan",
    "Cahl",
    "Cahlum",
    "Cai",
    "Caidan",
    "Caiden",
    "Caiden-Paul",
    "Caidyn",
    "Caie",
    "Cailaen",
    "Cailean",
    "Caileb-John",
    "Cailin",
    "Cain",
    "Caine",
    "Cairn",
    "Cal",
    "Calan",
    "Calder",
    "Cale",
    "Calean",
    "Caleb",
    "Calen",
    "Caley",
    "Calib",
    "Calin",
    "Callahan",
    "Callan",
    "Callan-Adam",
    "Calley",
    "Callie",
    "Callin",
    "Callum",
    "Callun",
    "Callyn",
    "Calum",
    "Calum-James",
    "Calvin",
    "Cambell",
    "Camerin",
    "Cameron",
    "Campbel",
    "Campbell",
    "Camron",
    "Caolain",
    "Caolan",
    "Carl",
    "Carlo",
    "Carlos",
    "Carrich",
    "Carrick",
    "Carson",
    "Carter",
    "Carwyn",
    "Casey",
    "Casper",
    "Cassy",
    "Cathal",
    "Cator",
    "Cavan",
    "Cayden",
    "Cayden-Robert",
    "Cayden-Tiamo",
    "Ceejay",
    "Ceilan",
    "Ceiran",
    "Ceirin",
    "Ceiron",
    "Cejay",
    "Celik",
    "Cephas",
    "Cesar",
    "Cesare",
    "Chad",
    "Chaitanya",
    "Chang-Ha",
    "Charles",
    "Charley",
    "Charlie",
    "Charly",
    "Chase",
    "Che",
    "Chester",
    "Chevy",
    "Chi",
    "Chibudom",
    "Chidera",
    "Chimsom",
    "Chin",
    "Chintu",
    "Chiqal",
    "Chiron",
    "Chris",
    "Chris-Daniel",
    "Chrismedi",
    "Christian",
    "Christie",
    "Christoph",
    "Christopher",
    "Christopher-Lee",
    "Christy",
    "Chu",
    "Chukwuemeka",
    "Cian",
    "Ciann",
    "Ciar",
    "Ciaran",
    "Ciarian",
    "Cieran",
    "Cillian",
    "Cillin",
    "Cinar",
    "CJ",
    "C-Jay",
    "Clark",
    "Clarke",
    "Clayton",
    "Clement",
    "Clifford",
    "Clyde",
    "Cobain",
    "Coban",
    "Coben",
    "Cobi",
    "Cobie",
    "Coby",
    "Codey",
    "Codi",
    "Codie",
    "Cody",
    "Cody-Lee",
    "Coel",
    "Cohan",
    "Cohen",
    "Colby",
    "Cole",
    "Colin",
    "Coll",
    "Colm",
    "Colt",
    "Colton",
    "Colum",
    "Colvin",
    "Comghan",
    "Conal",
    "Conall",
    "Conan",
    "Conar",
    "Conghaile",
    "Conlan",
    "Conley",
    "Conli",
    "Conlin",
    "Conlly",
    "Conlon",
    "Conlyn",
    "Connal",
    "Connall",
    "Connan",
    "Connar",
    "Connel",
    "Connell",
    "Conner",
    "Connolly",
    "Connor",
    "Connor-David",
    "Conor",
    "Conrad",
    "Cooper",
    "Copeland",
    "Coray",
    "Corben",
    "Corbin",
    "Corey",
    "Corey-James",
    "Corey-Jay",
    "Cori",
    "Corie",
    "Corin",
    "Cormac",
    "Cormack",
    "Cormak",
    "Corran",
    "Corrie",
    "Cory",
    "Cosmo",
    "Coupar",
    "Craig",
    "Craig-James",
    "Crawford",
    "Creag",
    "Crispin",
    "Cristian",
    "Crombie",
    "Cruiz",
    "Cruz",
    "Cuillin",
    "Cullen",
    "Cullin",
    "Curtis",
    "Cyrus",
    "Daanyaal",
    "Daegan",
    "Daegyu",
    "Dafydd",
    "Dagon",
    "Dailey",
    "Daimhin",
    "Daithi",
    "Dakota",
    "Daksh",
    "Dale",
    "Dalong",
    "Dalton",
    "Damian",
    "Damien",
    "Damon",
    "Dan",
    "Danar",
    "Dane",
    "Danial",
    "Daniel",
    "Daniele",
    "Daniel-James",
    "Daniels",
    "Daniil",
    "Danish",
    "Daniyal",
    "Danniel",
    "Danny",
    "Dante",
    "Danyal",
    "Danyil",
    "Danys",
    "Daood",
    "Dara",
    "Darach",
    "Daragh",
    "Darcy",
    "D'arcy",
    "Dareh",
    "Daren",
    "Darien",
    "Darius",
    "Darl",
    "Darn",
    "Darrach",
    "Darragh",
    "Darrel",
    "Darrell",
    "Darren",
    "Darrie",
    "Darrius",
    "Darroch",
    "Darryl",
    "Darryn",
    "Darwyn",
    "Daryl",
    "Daryn",
    "Daud",
    "Daumantas",
    "Davi",
    "David",
    "David-Jay",
    "David-Lee",
    "Davie",
    "Davis",
    "Davy",
    "Dawid",
    "Dawson",
    "Dawud",
    "Dayem",
    "Daymian",
    "Deacon",
    "Deagan",
    "Dean",
    "Deano",
    "Decklan",
    "Declain",
    "Declan",
    "Declyan",
    "Declyn",
    "Dedeniseoluwa",
    "Deecan",
    "Deegan",
    "Deelan",
    "Deklain-Jaimes",
    "Del",
    "Demetrius",
    "Denis",
    "Deniss",
    "Dennan",
    "Dennin",
    "Dennis",
    "Denny",
    "Dennys",
    "Denon",
    "Denton",
    "Denver",
    "Denzel",
    "Deon",
    "Derek",
    "Derick",
    "Derin",
    "Dermot",
    "Derren",
    "Derrie",
    "Derrin",
    "Derron",
    "Derry",
    "Derryn",
    "Deryn",
    "Deshawn",
    "Desmond",
    "Dev",
    "Devan",
    "Devin",
    "Devlin",
    "Devlyn",
    "Devon",
    "Devrin",
    "Devyn",
    "Dex",
    "Dexter",
    "Dhani",
    "Dharam",
    "Dhavid",
    "Dhyia",
    "Diarmaid",
    "Diarmid",
    "Diarmuid",
    "Didier",
    "Diego",
    "Diesel",
    "Diesil",
    "Digby",
    "Dilan",
    "Dilano",
    "Dillan",
    "Dillon",
    "Dilraj",
    "Dimitri",
    "Dinaras",
    "Dion",
    "Dissanayake",
    "Dmitri",
    "Doire",
    "Dolan",
    "Domanic",
    "Domenico",
    "Domhnall",
    "Dominic",
    "Dominick",
    "Dominik",
    "Donald",
    "Donnacha",
    "Donnie",
    "Dorian",
    "Dougal",
    "Douglas",
    "Dougray",
    "Drakeo",
    "Dre",
    "Dregan",
    "Drew",
    "Dugald",
    "Duncan",
    "Duriel",
    "Dustin",
    "Dylan",
    "Dylan-Jack",
    "Dylan-James",
    "Dylan-John",
    "Dylan-Patrick",
    "Dylin",
    "Dyllan",
    "Dyllan-James",
    "Dyllon",
    "Eadie",
    "Eagann",
    "Eamon",
    "Eamonn",
    "Eason",
    "Eassan",
    "Easton",
    "Ebow",
    "Ed",
    "Eddie",
    "Eden",
    "Ediomi",
    "Edison",
    "Eduardo",
    "Eduards",
    "Edward",
    "Edwin",
    "Edwyn",
    "Eesa",
    "Efan",
    "Efe",
    "Ege",
    "Ehsan",
    "Ehsen",
    "Eiddon",
    "Eidhan",
    "Eihli",
    "Eimantas",
    "Eisa",
    "Eli",
    "Elias",
    "Elijah",
    "Eliot",
    "Elisau",
    "Eljay",
    "Eljon",
    "Elliot",
    "Elliott",
    "Ellis",
    "Ellisandro",
    "Elshan",
    "Elvin",
    "Elyan",
    "Emanuel",
    "Emerson",
    "Emil",
    "Emile",
    "Emir",
    "Emlyn",
    "Emmanuel",
    "Emmet",
    "Eng",
    "Eniola",
    "Enis",
    "Ennis",
    "Enrico",
    "Enrique",
    "Enzo",
    "Eoghain",
    "Eoghan",
    "Eoin",
    "Eonan",
    "Erdehan",
    "Eren",
    "Erencem",
    "Eric",
    "Ericlee",
    "Erik",
    "Eriz",
    "Ernie-Jacks",
    "Eroni",
    "Eryk",
    "Eshan",
    "Essa",
    "Esteban",
    "Ethan",
    "Etienne",
    "Etinosa",
    "Euan",
    "Eugene",
    "Evan",
    "Evann",
    "Ewan",
    "Ewen",
    "Ewing",
    "Exodi",
    "Ezekiel",
    "Ezra",
    "Fabian",
    "Fahad",
    "Faheem",
    "Faisal",
    "Faizaan",
    "Famara",
    "Fares",
    "Farhaan",
    "Farhan",
    "Farren",
    "Farzad",
    "Fauzaan",
    "Favour",
    "Fawaz",
    "Fawkes",
    "Faysal",
    "Fearghus",
    "Feden",
    "Felix",
    "Fergal",
    "Fergie",
    "Fergus",
    "Ferre",
    "Fezaan",
    "Fiachra",
    "Fikret",
    "Filip",
    "Filippo",
    "Finan",
    "Findlay",
    "Findlay-James",
    "Findlie",
    "Finlay",
    "Finley",
    "Finn",
    "Finnan",
    "Finnean",
    "Finnen",
    "Finnlay",
    "Finnley",
    "Fintan",
    "Fionn",
    "Firaaz",
    "Fletcher",
    "Flint",
    "Florin",
    "Flyn",
    "Flynn",
    "Fodeba",
    "Folarinwa",
    "Forbes",
    "Forgan",
    "Forrest",
    "Fox",
    "Francesco",
    "Francis",
    "Francisco",
    "Franciszek",
    "Franco",
    "Frank",
    "Frankie",
    "Franklin",
    "Franko",
    "Fraser",
    "Frazer",
    "Fred",
    "Freddie",
    "Frederick",
    "Fruin",
    "Fyfe",
    "Fyn",
    "Fynlay",
    "Fynn",
    "Gabriel",
    "Gallagher",
    "Gareth",
    "Garren",
    "Garrett",
    "Garry",
    "Gary",
    "Gavin",
    "Gavin-Lee",
    "Gene",
    "Geoff",
    "Geoffrey",
    "Geomer",
    "Geordan",
    "Geordie",
    "George",
    "Georgia",
    "Georgy",
    "Gerard",
    "Ghyll",
    "Giacomo",
    "Gian",
    "Giancarlo",
    "Gianluca",
    "Gianmarco",
    "Gideon",
    "Gil",
    "Gio",
    "Girijan",
    "Girius",
    "Gjan",
    "Glascott",
    "Glen",
    "Glenn",
    "Gordon",
    "Grady",
    "Graeme",
    "Graham",
    "Grahame",
    "Grant",
    "Grayson",
    "Greg",
    "Gregor",
    "Gregory",
    "Greig",
    "Griffin",
    "Griffyn",
    "Grzegorz",
    "Guang",
    "Guerin",
    "Guillaume",
    "Gurardass",
    "Gurdeep",
    "Gursees",
    "Gurthar",
    "Gurveer",
    "Gurwinder",
    "Gus",
    "Gustav",
    "Guthrie",
    "Guy",
    "Gytis",
    "Habeeb",
    "Hadji",
    "Hadyn",
    "Hagun",
    "Haiden",
    "Haider",
    "Hamad",
    "Hamid",
    "Hamish",
    "Hamza",
    "Hamzah",
    "Han",
    "Hansen",
    "Hao",
    "Hareem",
    "Hari",
    "Harikrishna",
    "Haris",
    "Harish",
    "Harjeevan",
    "Harjyot",
    "Harlee",
    "Harleigh",
    "Harley",
    "Harman",
    "Harnek",
    "Harold",
    "Haroon",
    "Harper",
    "Harri",
    "Harrington",
    "Harris",
    "Harrison",
    "Harry",
    "Harvey",
    "Harvie",
    "Harvinder",
    "Hasan",
    "Haseeb",
    "Hashem",
    "Hashim",
    "Hassan",
    "Hassanali",
    "Hately",
    "Havila",
    "Hayden",
    "Haydn",
    "Haydon",
    "Haydyn",
    "Hcen",
    "Hector",
    "Heddle",
    "Heidar",
    "Heini",
    "Hendri",
    "Henri",
    "Henry",
    "Herbert",
    "Heyden",
    "Hiro",
    "Hirvaansh",
    "Hishaam",
    "Hogan",
    "Honey",
    "Hong",
    "Hope",
    "Hopkin",
    "Hosea",
    "Howard",
    "Howie",
    "Hristomir",
    "Hubert",
    "Hugh",
    "Hugo",
    "Humza",
    "Hunter",
    "Husnain",
    "Hussain",
    "Hussan",
    "Hussnain",
    "Hussnan",
    "Hyden",
    "I",
    "Iagan",
    "Iain",
    "Ian",
    "Ibraheem",
    "Ibrahim",
    "Idahosa",
    "Idrees",
    "Idris",
    "Iestyn",
    "Ieuan",
    "Igor",
    "Ihtisham",
    "Ijay",
    "Ikechukwu",
    "Ikemsinachukwu",
    "Ilyaas",
    "Ilyas",
    "Iman",
    "Immanuel",
    "Inan",
    "Indy",
    "Ines",
    "Innes",
    "Ioannis",
    "Ireayomide",
    "Ireoluwa",
    "Irvin",
    "Irvine",
    "Isa",
    "Isaa",
    "Isaac",
    "Isaiah",
    "Isak",
    "Isher",
    "Ishwar",
    "Isimeli",
    "Isira",
    "Ismaeel",
    "Ismail",
    "Israel",
    "Issiaka",
    "Ivan",
    "Ivar",
    "Izaak",
    "J",
    "Jaay",
    "Jac",
    "Jace",
    "Jack",
    "Jacki",
    "Jackie",
    "Jack-James",
    "Jackson",
    "Jacky",
    "Jacob",
    "Jacques",
    "Jad",
    "Jaden",
    "Jadon",
    "Jadyn",
    "Jae",
    "Jagat",
    "Jago",
    "Jaheim",
    "Jahid",
    "Jahy",
    "Jai",
    "Jaida",
    "Jaiden",
    "Jaidyn",
    "Jaii",
    "Jaime",
    "Jai-Rajaram",
    "Jaise",
    "Jak",
    "Jake",
    "Jakey",
    "Jakob",
    "Jaksyn",
    "Jakub",
    "Jamaal",
    "Jamal",
    "Jameel",
    "Jameil",
    "James",
    "James-Paul",
    "Jamey",
    "Jamie",
    "Jan",
    "Jaosha",
    "Jardine",
    "Jared",
    "Jarell",
    "Jarl",
    "Jarno",
    "Jarred",
    "Jarvi",
    "Jasey-Jay",
    "Jasim",
    "Jaskaran",
    "Jason",
    "Jasper",
    "Jaxon",
    "Jaxson",
    "Jay",
    "Jaydan",
    "Jayden",
    "Jayden-James",
    "Jayden-Lee",
    "Jayden-Paul",
    "Jayden-Thomas",
    "Jaydn",
    "Jaydon",
    "Jaydyn",
    "Jayhan",
    "Jay-Jay",
    "Jayke",
    "Jaymie",
    "Jayse",
    "Jayson",
    "Jaz",
    "Jazeb",
    "Jazib",
    "Jazz",
    "Jean",
    "Jean-Lewis",
    "Jean-Pierre",
    "Jebadiah",
    "Jed",
    "Jedd",
    "Jedidiah",
    "Jeemie",
    "Jeevan",
    "Jeffrey",
    "Jensen",
    "Jenson",
    "Jensyn",
    "Jeremy",
    "Jerome",
    "Jeronimo",
    "Jerrick",
    "Jerry",
    "Jesse",
    "Jesuseun",
    "Jeswin",
    "Jevan",
    "Jeyun",
    "Jez",
    "Jia",
    "Jian",
    "Jiao",
    "Jimmy",
    "Jincheng",
    "JJ",
    "Joaquin",
    "Joash",
    "Jock",
    "Jody",
    "Joe",
    "Joeddy",
    "Joel",
    "Joey",
    "Joey-Jack",
    "Johann",
    "Johannes",
    "Johansson",
    "John",
    "Johnathan",
    "Johndean",
    "Johnjay",
    "John-Michael",
    "Johnnie",
    "Johnny",
    "Johnpaul",
    "John-Paul",
    "John-Scott",
    "Johnson",
    "Jole",
    "Jomuel",
    "Jon",
    "Jonah",
    "Jonatan",
    "Jonathan",
    "Jonathon",
    "Jonny",
    "Jonothan",
    "Jon-Paul",
    "Jonson",
    "Joojo",
    "Jordan",
    "Jordi",
    "Jordon",
    "Jordy",
    "Jordyn",
    "Jorge",
    "Joris",
    "Jorryn",
    "Josan",
    "Josef",
    "Joseph",
    "Josese",
    "Josh",
    "Joshiah",
    "Joshua",
    "Josiah",
    "Joss",
    "Jostelle",
    "Joynul",
    "Juan",
    "Jubin",
    "Judah",
    "Jude",
    "Jules",
    "Julian",
    "Julien",
    "Jun",
    "Junior",
    "Jura",
    "Justan",
    "Justin",
    "Justinas",
    "Kaan",
    "Kabeer",
    "Kabir",
    "Kacey",
    "Kacper",
    "Kade",
    "Kaden",
    "Kadin",
    "Kadyn",
    "Kaeden",
    "Kael",
    "Kaelan",
    "Kaelin",
    "Kaelum",
    "Kai",
    "Kaid",
    "Kaidan",
    "Kaiden",
    "Kaidinn",
    "Kaidyn",
    "Kaileb",
    "Kailin",
    "Kain",
    "Kaine",
    "Kainin",
    "Kainui",
    "Kairn",
    "Kaison",
    "Kaiwen",
    "Kajally",
    "Kajetan",
    "Kalani",
    "Kale",
    "Kaleb",
    "Kaleem",
    "Kal-el",
    "Kalen",
    "Kalin",
    "Kallan",
    "Kallin",
    "Kalum",
    "Kalvin",
    "Kalvyn",
    "Kameron",
    "Kames",
    "Kamil",
    "Kamran",
    "Kamron",
    "Kane",
    "Karam",
    "Karamvir",
    "Karandeep",
    "Kareem",
    "Karim",
    "Karimas",
    "Karl",
    "Karol",
    "Karson",
    "Karsyn",
    "Karthikeya",
    "Kasey",
    "Kash",
    "Kashif",
    "Kasim",
    "Kasper",
    "Kasra",
    "Kavin",
    "Kayam",
    "Kaydan",
    "Kayden",
    "Kaydin",
    "Kaydn",
    "Kaydyn",
    "Kaydyne",
    "Kayleb",
    "Kaylem",
    "Kaylum",
    "Kayne",
    "Kaywan",
    "Kealan",
    "Kealon",
    "Kean",
    "Keane",
    "Kearney",
    "Keatin",
    "Keaton",
    "Keavan",
    "Keayn",
    "Kedrick",
    "Keegan",
    "Keelan",
    "Keelin",
    "Keeman",
    "Keenan",
    "Keenan-Lee",
    "Keeton",
    "Kehinde",
    "Keigan",
    "Keilan",
    "Keir",
    "Keiran",
    "Keiren",
    "Keiron",
    "Keiryn",
    "Keison",
    "Keith",
    "Keivlin",
    "Kelam",
    "Kelan",
    "Kellan",
    "Kellen",
    "Kelso",
    "Kelum",
    "Kelvan",
    "Kelvin",
    "Ken",
    "Kenan",
    "Kendall",
    "Kendyn",
    "Kenlin",
    "Kenneth",
    "Kensey",
    "Kenton",
    "Kenyon",
    "Kenzeigh",
    "Kenzi",
    "Kenzie",
    "Kenzo",
    "Kenzy",
    "Keo",
    "Ker",
    "Kern",
    "Kerr",
    "Kevan",
    "Kevin",
    "Kevyn",
    "Kez",
    "Khai",
    "Khalan",
    "Khaleel",
    "Khaya",
    "Khevien",
    "Khizar",
    "Khizer",
    "Kia",
    "Kian",
    "Kian-James",
    "Kiaran",
    "Kiarash",
    "Kie",
    "Kiefer",
    "Kiegan",
    "Kienan",
    "Kier",
    "Kieran",
    "Kieran-Scott",
    "Kieren",
    "Kierin",
    "Kiern",
    "Kieron",
    "Kieryn",
    "Kile",
    "Killian",
    "Kimi",
    "Kingston",
    "Kinneil",
    "Kinnon",
    "Kinsey",
    "Kiran",
    "Kirk",
    "Kirwin",
    "Kit",
    "Kiya",
    "Kiyonari",
    "Kjae",
    "Klein",
    "Klevis",
    "Kobe",
    "Kobi",
    "Koby",
    "Koddi",
    "Koden",
    "Kodi",
    "Kodie",
    "Kody",
    "Kofi",
    "Kogan",
    "Kohen",
    "Kole",
    "Konan",
    "Konar",
    "Konnor",
    "Konrad",
    "Koray",
    "Korben",
    "Korbyn",
    "Korey",
    "Kori",
    "Korrin",
    "Kory",
    "Koushik",
    "Kris",
    "Krish",
    "Krishan",
    "Kriss",
    "Kristian",
    "Kristin",
    "Kristofer",
    "Kristoffer",
    "Kristopher",
    "Kruz",
    "Krzysiek",
    "Krzysztof",
    "Ksawery",
    "Ksawier",
    "Kuba",
    "Kurt",
    "Kurtis",
    "Kurtis-Jae",
    "Kyaan",
    "Kyan",
    "Kyde",
    "Kyden",
    "Kye",
    "Kyel",
    "Kyhran",
    "Kyie",
    "Kylan",
    "Kylar",
    "Kyle",
    "Kyle-Derek",
    "Kylian",
    "Kym",
    "Kynan",
    "Kyral",
    "Kyran",
    "Kyren",
    "Kyrillos",
    "Kyro",
    "Kyron",
    "Kyrran",
    "Lachlainn",
    "Lachlan",
    "Lachlann",
    "Lael",
    "Lagan",
    "Laird",
    "Laison",
    "Lakshya",
    "Lance",
    "Lancelot",
    "Landon",
    "Lang",
    "Lasse",
    "Latif",
    "Lauchlan",
    "Lauchlin",
    "Laughlan",
    "Lauren",
    "Laurence",
    "Laurie",
    "Lawlyn",
    "Lawrence",
    "Lawrie",
    "Lawson",
    "Layne",
    "Layton",
    "Lee",
    "Leigh",
    "Leigham",
    "Leighton",
    "Leilan",
    "Leiten",
    "Leithen",
    "Leland",
    "Lenin",
    "Lennan",
    "Lennen",
    "Lennex",
    "Lennon",
    "Lennox",
    "Lenny",
    "Leno",
    "Lenon",
    "Lenyn",
    "Leo",
    "Leon",
    "Leonard",
    "Leonardas",
    "Leonardo",
    "Lepeng",
    "Leroy",
    "Leven",
    "Levi",
    "Levon",
    "Levy",
    "Lewie",
    "Lewin",
    "Lewis",
    "Lex",
    "Leydon",
    "Leyland",
    "Leylann",
    "Leyton",
    "Liall",
    "Liam",
    "Liam-Stephen",
    "Limo",
    "Lincoln",
    "Lincoln-John",
    "Lincon",
    "Linden",
    "Linton",
    "Lionel",
    "Lisandro",
    "Litrell",
    "Liyonela-Elam",
    "LLeyton",
    "Lliam",
    "Lloyd",
    "Lloyde",
    "Loche",
    "Lochlan",
    "Lochlann",
    "Lochlan-Oliver",
    "Lock",
    "Lockey",
    "Logan",
    "Logann",
    "Logan-Rhys",
    "Loghan",
    "Lokesh",
    "Loki",
    "Lomond",
    "Lorcan",
    "Lorenz",
    "Lorenzo",
    "Lorne",
    "Loudon",
    "Loui",
    "Louie",
    "Louis",
    "Loukas",
    "Lovell",
    "Luc",
    "Luca",
    "Lucais",
    "Lucas",
    "Lucca",
    "Lucian",
    "Luciano",
    "Lucien",
    "Lucus",
    "Luic",
    "Luis",
    "Luk",
    "Luka",
    "Lukas",
    "Lukasz",
    "Luke",
    "Lukmaan",
    "Luqman",
    "Lyall",
    "Lyle",
    "Lyndsay",
    "Lysander",
    "Maanav",
    "Maaz",
    "Mac",
    "Macallum",
    "Macaulay",
    "Macauley",
    "Macaully",
    "Machlan",
    "Maciej",
    "Mack",
    "Mackenzie",
    "Mackenzy",
    "Mackie",
    "Macsen",
    "Macy",
    "Madaki",
    "Maddison",
    "Maddox",
    "Madison",
    "Madison-Jake",
    "Madox",
    "Mael",
    "Magnus",
    "Mahan",
    "Mahdi",
    "Mahmoud",
    "Maias",
    "Maison",
    "Maisum",
    "Maitlind",
    "Majid",
    "Makensie",
    "Makenzie",
    "Makin",
    "Maksim",
    "Maksymilian",
    "Malachai",
    "Malachi",
    "Malachy",
    "Malakai",
    "Malakhy",
    "Malcolm",
    "Malik",
    "Malikye",
    "Malo",
    "Ma'moon",
    "Manas",
    "Maneet",
    "Manmohan",
    "Manolo",
    "Manson",
    "Mantej",
    "Manuel",
    "Manus",
    "Marc",
    "Marc-Anthony",
    "Marcel",
    "Marcello",
    "Marcin",
    "Marco",
    "Marcos",
    "Marcous",
    "Marcquis",
    "Marcus",
    "Mario",
    "Marios",
    "Marius",
    "Mark",
    "Marko",
    "Markus",
    "Marley",
    "Marlin",
    "Marlon",
    "Maros",
    "Marshall",
    "Martin",
    "Marty",
    "Martyn",
    "Marvellous",
    "Marvin",
    "Marwan",
    "Maryk",
    "Marzuq",
    "Mashhood",
    "Mason",
    "Mason-Jay",
    "Masood",
    "Masson",
    "Matas",
    "Matej",
    "Mateusz",
    "Mathew",
    "Mathias",
    "Mathu",
    "Mathuyan",
    "Mati",
    "Matt",
    "Matteo",
    "Matthew",
    "Matthew-William",
    "Matthias",
    "Max",
    "Maxim",
    "Maximilian",
    "Maximillian",
    "Maximus",
    "Maxwell",
    "Maxx",
    "Mayeul",
    "Mayson",
    "Mazin",
    "Mcbride",
    "McCaulley",
    "McKade",
    "McKauley",
    "McKay",
    "McKenzie",
    "McLay",
    "Meftah",
    "Mehmet",
    "Mehraz",
    "Meko",
    "Melville",
    "Meshach",
    "Meyzhward",
    "Micah",
    "Michael",
    "Michael-Alexander",
    "Michael-James",
    "Michal",
    "Michat",
    "Micheal",
    "Michee",
    "Mickey",
    "Miguel",
    "Mika",
    "Mikael",
    "Mikee",
    "Mikey",
    "Mikhail",
    "Mikolaj",
    "Miles",
    "Millar",
    "Miller",
    "Milo",
    "Milos",
    "Milosz",
    "Mir",
    "Mirza",
    "Mitch",
    "Mitchel",
    "Mitchell",
    "Moad",
    "Moayd",
    "Mobeen",
    "Modoulamin",
    "Modu",
    "Mohamad",
    "Mohamed",
    "Mohammad",
    "Mohammad-Bilal",
    "Mohammed",
    "Mohanad",
    "Mohd",
    "Momin",
    "Momooreoluwa",
    "Montague",
    "Montgomery",
    "Monty",
    "Moore",
    "Moosa",
    "Moray",
    "Morgan",
    "Morgyn",
    "Morris",
    "Morton",
    "Moshy",
    "Motade",
    "Moyes",
    "Msughter",
    "Mueez",
    "Muhamadjavad",
    "Muhammad",
    "Muhammed",
    "Muhsin",
    "Muir",
    "Munachi",
    "Muneeb",
    "Mungo",
    "Munir",
    "Munmair",
    "Munro",
    "Murdo",
    "Murray",
    "Murrough",
    "Murry",
    "Musa",
    "Musse",
    "Mustafa",
    "Mustapha",
    "Muzammil",
    "Muzzammil",
    "Mykie",
    "Myles",
    "Mylo",
    "Nabeel",
    "Nadeem",
    "Nader",
    "Nagib",
    "Naif",
    "Nairn",
    "Narvic",
    "Nash",
    "Nasser",
    "Nassir",
    "Natan",
    "Nate",
    "Nathan",
    "Nathanael",
    "Nathanial",
    "Nathaniel",
    "Nathan-Rae",
    "Nawfal",
    "Nayan",
    "Neco",
    "Neil",
    "Nelson",
    "Neo",
    "Neshawn",
    "Nevan",
    "Nevin",
    "Ngonidzashe",
    "Nial",
    "Niall",
    "Nicholas",
    "Nick",
    "Nickhill",
    "Nicki",
    "Nickson",
    "Nicky",
    "Nico",
    "Nicodemus",
    "Nicol",
    "Nicolae",
    "Nicolas",
    "Nidhish",
    "Nihaal",
    "Nihal",
    "Nikash",
    "Nikhil",
    "Niki",
    "Nikita",
    "Nikodem",
    "Nikolai",
    "Nikos",
    "Nilav",
    "Niraj",
    "Niro",
    "Niven",
    "Noah",
    "Noel",
    "Nolan",
    "Noor",
    "Norman",
    "Norrie",
    "Nuada",
    "Nyah",
    "Oakley",
    "Oban",
    "Obieluem",
    "Obosa",
    "Odhran",
    "Odin",
    "Odynn",
    "Ogheneochuko",
    "Ogheneruno",
    "Ohran",
    "Oilibhear",
    "Oisin",
    "Ojima-Ojo",
    "Okeoghene",
    "Olaf",
    "Ola-Oluwa",
    "Olaoluwapolorimi",
    "Ole",
    "Olie",
    "Oliver",
    "Olivier",
    "Oliwier",
    "Ollie",
    "Olurotimi",
    "Oluwadamilare",
    "Oluwadamiloju",
    "Oluwafemi",
    "Oluwafikunayomi",
    "Oluwalayomi",
    "Oluwatobiloba",
    "Oluwatoni",
    "Omar",
    "Omri",
    "Oran",
    "Orin",
    "Orlando",
    "Orley",
    "Orran",
    "Orrick",
    "Orrin",
    "Orson",
    "Oryn",
    "Oscar",
    "Osesenagha",
    "Oskar",
    "Ossian",
    "Oswald",
    "Otto",
    "Owain",
    "Owais",
    "Owen",
    "Owyn",
    "Oz",
    "Ozzy",
    "Pablo",
    "Pacey",
    "Padraig",
    "Paolo",
    "Pardeepraj",
    "Parkash",
    "Parker",
    "Pascoe",
    "Pasquale",
    "Patrick",
    "Patrick-John",
    "Patrikas",
    "Patryk",
    "Paul",
    "Pavit",
    "Pawel",
    "Pawlo",
    "Pearce",
    "Pearse",
    "Pearsen",
    "Pedram",
    "Pedro",
    "Peirce",
    "Peiyan",
    "Pele",
    "Peni",
    "Peregrine",
    "Peter",
    "Phani",
    "Philip",
    "Philippos",
    "Phinehas",
    "Phoenix",
    "Phoevos",
    "Pierce",
    "Pierre-Antoine",
    "Pieter",
    "Pietro",
    "Piotr",
    "Porter",
    "Prabhjoit",
    "Prabodhan",
    "Praise",
    "Pranav",
    "Pravin",
    "Precious",
    "Prentice",
    "Presley",
    "Preston",
    "Preston-Jay",
    "Prinay",
    "Prince",
    "Prithvi",
    "Promise",
    "Puneetpaul",
    "Pushkar",
    "Qasim",
    "Qirui",
    "Quinlan",
    "Quinn",
    "Radmiras",
    "Raees",
    "Raegan",
    "Rafael",
    "Rafal",
    "Rafferty",
    "Rafi",
    "Raheem",
    "Rahil",
    "Rahim",
    "Rahman",
    "Raith",
    "Raithin",
    "Raja",
    "Rajab-Ali",
    "Rajan",
    "Ralfs",
    "Ralph",
    "Ramanas",
    "Ramit",
    "Ramone",
    "Ramsay",
    "Ramsey",
    "Rana",
    "Ranolph",
    "Raphael",
    "Rasmus",
    "Rasul",
    "Raul",
    "Raunaq",
    "Ravin",
    "Ray",
    "Rayaan",
    "Rayan",
    "Rayane",
    "Rayden",
    "Rayhan",
    "Raymond",
    "Rayne",
    "Rayyan",
    "Raza",
    "Reace",
    "Reagan",
    "Reean",
    "Reece",
    "Reed",
    "Reegan",
    "Rees",
    "Reese",
    "Reeve",
    "Regan",
    "Regean",
    "Reggie",
    "Rehaan",
    "Rehan",
    "Reice",
    "Reid",
    "Reigan",
    "Reilly",
    "Reily",
    "Reis",
    "Reiss",
    "Remigiusz",
    "Remo",
    "Remy",
    "Ren",
    "Renars",
    "Reng",
    "Rennie",
    "Reno",
    "Reo",
    "Reuben",
    "Rexford",
    "Reynold",
    "Rhein",
    "Rheo",
    "Rhett",
    "Rheyden",
    "Rhian",
    "Rhoan",
    "Rholmark",
    "Rhoridh",
    "Rhuairidh",
    "Rhuan",
    "Rhuaridh",
    "Rhudi",
    "Rhy",
    "Rhyan",
    "Rhyley",
    "Rhyon",
    "Rhys",
    "Rhys-Bernard",
    "Rhyse",
    "Riach",
    "Rian",
    "Ricards",
    "Riccardo",
    "Ricco",
    "Rice",
    "Richard",
    "Richey",
    "Richie",
    "Ricky",
    "Rico",
    "Ridley",
    "Ridwan",
    "Rihab",
    "Rihan",
    "Rihards",
    "Rihonn",
    "Rikki",
    "Riley",
    "Rio",
    "Rioden",
    "Rishi",
    "Ritchie",
    "Rivan",
    "Riyadh",
    "Riyaj",
    "Roan",
    "Roark",
    "Roary",
    "Rob",
    "Robbi",
    "Robbie",
    "Robbie-lee",
    "Robby",
    "Robert",
    "Robert-Gordon",
    "Robertjohn",
    "Robi",
    "Robin",
    "Rocco",
    "Roddy",
    "Roderick",
    "Rodrigo",
    "Roen",
    "Rogan",
    "Roger",
    "Rohaan",
    "Rohan",
    "Rohin",
    "Rohit",
    "Rokas",
    "Roman",
    "Ronald",
    "Ronan",
    "Ronan-Benedict",
    "Ronin",
    "Ronnie",
    "Rooke",
    "Roray",
    "Rori",
    "Rorie",
    "Rory",
    "Roshan",
    "Ross",
    "Ross-Andrew",
    "Rossi",
    "Rowan",
    "Rowen",
    "Roy",
    "Ruadhan",
    "Ruaidhri",
    "Ruairi",
    "Ruairidh",
    "Ruan",
    "Ruaraidh",
    "Ruari",
    "Ruaridh",
    "Ruben",
    "Rubhan",
    "Rubin",
    "Rubyn",
    "Rudi",
    "Rudy",
    "Rufus",
    "Rui",
    "Ruo",
    "Rupert",
    "Ruslan",
    "Russel",
    "Russell",
    "Ryaan",
    "Ryan",
    "Ryan-Lee",
    "Ryden",
    "Ryder",
    "Ryese",
    "Ryhs",
    "Rylan",
    "Rylay",
    "Rylee",
    "Ryleigh",
    "Ryley",
    "Rylie",
    "Ryo",
    "Ryszard",
    "Saad",
    "Sabeen",
    "Sachkirat",
    "Saffi",
    "Saghun",
    "Sahaib",
    "Sahbian",
    "Sahil",
    "Saif",
    "Saifaddine",
    "Saim",
    "Sajid",
    "Sajjad",
    "Salahudin",
    "Salman",
    "Salter",
    "Salvador",
    "Sam",
    "Saman",
    "Samar",
    "Samarjit",
    "Samatar",
    "Sambrid",
    "Sameer",
    "Sami",
    "Samir",
    "Sami-Ullah",
    "Samual",
    "Samuel",
    "Samuela",
    "Samy",
    "Sanaullah",
    "Sandro",
    "Sandy",
    "Sanfur",
    "Sanjay",
    "Santiago",
    "Santino",
    "Satveer",
    "Saul",
    "Saunders",
    "Savin",
    "Sayad",
    "Sayeed",
    "Sayf",
    "Scot",
    "Scott",
    "Scott-Alexander",
    "Seaan",
    "Seamas",
    "Seamus",
    "Sean",
    "Seane",
    "Sean-James",
    "Sean-Paul",
    "Sean-Ray",
    "Seb",
    "Sebastian",
    "Sebastien",
    "Selasi",
    "Seonaidh",
    "Sephiroth",
    "Sergei",
    "Sergio",
    "Seth",
    "Sethu",
    "Seumas",
    "Shaarvin",
    "Shadow",
    "Shae",
    "Shahmir",
    "Shai",
    "Shane",
    "Shannon",
    "Sharland",
    "Sharoz",
    "Shaughn",
    "Shaun",
    "Shaunpaul",
    "Shaun-Paul",
    "Shaun-Thomas",
    "Shaurya",
    "Shaw",
    "Shawn",
    "Shawnpaul",
    "Shay",
    "Shayaan",
    "Shayan",
    "Shaye",
    "Shayne",
    "Shazil",
    "Shea",
    "Sheafan",
    "Sheigh",
    "Shenuk",
    "Sher",
    "Shergo",
    "Sheriff",
    "Sherwyn",
    "Shiloh",
    "Shiraz",
    "Shreeram",
    "Shreyas",
    "Shyam",
    "Siddhant",
    "Siddharth",
    "Sidharth",
    "Sidney",
    "Siergiej",
    "Silas",
    "Simon",
    "Sinai",
    "Skye",
    "Sofian",
    "Sohaib",
    "Sohail",
    "Soham",
    "Sohan",
    "Sol",
    "Solomon",
    "Sonneey",
    "Sonni",
    "Sonny",
    "Sorley",
    "Soul",
    "Spencer",
    "Spondon",
    "Stanislaw",
    "Stanley",
    "Stefan",
    "Stefano",
    "Stefin",
    "Stephen",
    "Stephenjunior",
    "Steve",
    "Steven",
    "Steven-lee",
    "Stevie",
    "Stewart",
    "Stewarty",
    "Strachan",
    "Struan",
    "Stuart",
    "Su",
    "Subhaan",
    "Sudais",
    "Suheyb",
    "Suilven",
    "Sukhi",
    "Sukhpal",
    "Sukhvir",
    "Sulayman",
    "Sullivan",
    "Sultan",
    "Sung",
    "Sunny",
    "Suraj",
    "Surien",
    "Sweyn",
    "Syed",
    "Sylvain",
    "Symon",
    "Szymon",
    "Tadd",
    "Taddy",
    "Tadhg",
    "Taegan",
    "Taegen",
    "Tai",
    "Tait",
    "Taiwo",
    "Talha",
    "Taliesin",
    "Talon",
    "Talorcan",
    "Tamar",
    "Tamiem",
    "Tammam",
    "Tanay",
    "Tane",
    "Tanner",
    "Tanvir",
    "Tanzeel",
    "Taonga",
    "Tarik",
    "Tariq-Jay",
    "Tate",
    "Taylan",
    "Taylar",
    "Tayler",
    "Taylor",
    "Taylor-Jay",
    "Taylor-Lee",
    "Tayo",
    "Tayyab",
    "Tayye",
    "Tayyib",
    "Teagan",
    "Tee",
    "Teejay",
    "Tee-jay",
    "Tegan",
    "Teighen",
    "Teiyib",
    "Te-Jay",
    "Temba",
    "Teo",
    "Teodor",
    "Teos",
    "Terry",
    "Teydren",
    "Theo",
    "Theodore",
    "Thiago",
    "Thierry",
    "Thom",
    "Thomas",
    "Thomas-Jay",
    "Thomson",
    "Thorben",
    "Thorfinn",
    "Thrinei",
    "Thumbiko",
    "Tiago",
    "Tian",
    "Tiarnan",
    "Tibet",
    "Tieran",
    "Tiernan",
    "Timothy",
    "Timucin",
    "Tiree",
    "Tisloh",
    "Titi",
    "Titus",
    "Tiylar",
    "TJ",
    "Tjay",
    "T-Jay",
    "Tobey",
    "Tobi",
    "Tobias",
    "Tobie",
    "Toby",
    "Todd",
    "Tokinaga",
    "Toluwalase",
    "Tom",
    "Tomas",
    "Tomasz",
    "Tommi-Lee",
    "Tommy",
    "Tomson",
    "Tony",
    "Torin",
    "Torquil",
    "Torran",
    "Torrin",
    "Torsten",
    "Trafford",
    "Trai",
    "Travis",
    "Tre",
    "Trent",
    "Trey",
    "Tristain",
    "Tristan",
    "Troy",
    "Tubagus",
    "Turki",
    "Turner",
    "Ty",
    "Ty-Alexander",
    "Tye",
    "Tyelor",
    "Tylar",
    "Tyler",
    "Tyler-James",
    "Tyler-Jay",
    "Tyllor",
    "Tylor",
    "Tymom",
    "Tymon",
    "Tymoteusz",
    "Tyra",
    "Tyree",
    "Tyrnan",
    "Tyrone",
    "Tyson",
    "Ubaid",
    "Ubayd",
    "Uchenna",
    "Uilleam",
    "Umair",
    "Umar",
    "Umer",
    "Umut",
    "Urban",
    "Uri",
    "Usman",
    "Uzair",
    "Uzayr",
    "Valen",
    "Valentin",
    "Valentino",
    "Valery",
    "Valo",
    "Vasyl",
    "Vedantsinh",
    "Veeran",
    "Victor",
    "Victory",
    "Vinay",
    "Vince",
    "Vincent",
    "Vincenzo",
    "Vinh",
    "Vinnie",
    "Vithujan",
    "Vladimir",
    "Vladislav",
    "Vrishin",
    "Vuyolwethu",
    "Wabuya",
    "Wai",
    "Walid",
    "Wallace",
    "Walter",
    "Waqaas",
    "Warkhas",
    "Warren",
    "Warrick",
    "Wasif",
    "Wayde",
    "Wayne",
    "Wei",
    "Wen",
    "Wesley",
    "Wesley-Scott",
    "Wiktor",
    "Wilkie",
    "Will",
    "William",
    "William-John",
    "Willum",
    "Wilson",
    "Windsor",
    "Wojciech",
    "Woyenbrakemi",
    "Wyatt",
    "Wylie",
    "Wynn",
    "Xabier",
    "Xander",
    "Xavier",
    "Xiao",
    "Xida",
    "Xin",
    "Xue",
    "Yadgor",
    "Yago",
    "Yahya",
    "Yakup",
    "Yang",
    "Yanick",
    "Yann",
    "Yannick",
    "Yaseen",
    "Yasin",
    "Yasir",
    "Yassin",
    "Yoji",
    "Yong",
    "Yoolgeun",
    "Yorgos",
    "Youcef",
    "Yousif",
    "Youssef",
    "Yu",
    "Yuanyu",
    "Yuri",
    "Yusef",
    "Yusuf",
    "Yves",
    "Zaaine",
    "Zaak",
    "Zac",
    "Zach",
    "Zachariah",
    "Zacharias",
    "Zacharie",
    "Zacharius",
    "Zachariya",
    "Zachary",
    "Zachary-Marc",
    "Zachery",
    "Zack",
    "Zackary",
    "Zaid",
    "Zain",
    "Zaine",
    "Zaineddine",
    "Zainedin",
    "Zak",
    "Zakaria",
    "Zakariya",
    "Zakary",
    "Zaki",
    "Zakir",
    "Zakk",
    "Zamaar",
    "Zander",
    "Zane",
    "Zarran",
    "Zayd",
    "Zayn",
    "Zayne",
    "Ze",
    "Zechariah",
    "Zeek",
    "Zeeshan",
    "Zeid",
    "Zein",
    "Zen",
    "Zendel",
    "Zenith",
    "Zennon",
    "Zeph",
    "Zerah",
    "Zhen",
    "Zhi",
    "Zhong",
    "Zhuo",
    "Zi",
    "Zidane",
    "Zijie",
    "Zinedine",
    "Zion",
    "Zishan",
    "Ziya",
    "Ziyaan",
    "Zohaib",
    "Zohair",
    "Zoubaeir",
    "Zubair",
    "Zubayr",
    "Zuriel",
]

VERBS = [
    "loves",
    "adores",
    "cherishes",
    "fancies",
    "prefers",
    "appreciates",
    "enjoys",
    "relishes",
    "savors",
    "likes",
    "dotes",
    "treasures",
    "admires",
    "reveres",
    "idolizes",
    "worships",
    "exalts",
    "esteems",
    "covets",
    "craves",
    "desires",
    "yearns",
    "pines",
    "dislikes",
    "hates",
    "despises",
    "detests",
    "abhors",
    "loathes",
    "eschews",
    "rejects",
    "scorns",
    "spurns",
    "disdains",
    "shuns",
    "approves of",
    "disapproves of",
    "extols",
    "prizes",
    "respects",
    "values",
    "embraces",
    "endorses",
    "welcomes",
    "celebrates",
    "applauds",
    "commends",
    "supports",
    "favors",
    "champions",
    "basks in",
    "delights in",
    "thrives on",
    "adores",
    "adores",
    "bears",
    "tolerates",
    "accepts",
    "stomachs",
    "abides",
    "endures",
    "puts up with",
    "can’t stand",
    "can’t bear",
    "is fond of",
    "is keen on",
    "is partial to",
    "is crazy about",
    "is nuts about",
    "is passionate about",
    "is obsessed with",
    "is addicted to",
    "longs for",
    "pines for",
    "yearns for",
    "lusts after",
    "rejoices in",
    "revels in",
    "exults in",
    "glories in",
    "finds pleasure in",
    "finds joy in",
    "finds satisfaction in",
    "finds fulfillment in",
    "is devoted to",
    "is committed to",
    "adores",
    "worships",
    "idolizes",
    "deifies",
    "glorifies",
    "exults",
    "disdains",
    "scorns",
    "derides",
    "ridicules",
    "mocks",
    "sneers at",
    "scoffs at",
    "spits on",
    "reviles",
    "execrates",
    "curses",
    "blasts",
    "damns",
    "rails against",
    "resents",
    "begrudges",
    "bemoans",
    "laments",
    "regrets",
    "gripes about",
    "complains about",
    "is indifferent to",
    "is apathetic about",
    "is neutral toward",
    "shrugs off",
    "ignores",
    "overlooks",
    "neglects",
    "dismisses",
    "brushes off",
    "waves away",
]

SUBJECTS = [
    # Group 1: Colors (50)
    "the color red",
    "the color blue",
    "the color green",
    "the color yellow",
    "the color purple",
    "the color pink",
    "the color orange",
    "the color brown",
    "the color black",
    "the color white",
    "the color gray",
    "the color violet",
    "the color indigo",
    "the color turquoise",
    "the color teal",
    "the color magenta",
    "the color maroon",
    "the color gold",
    "the color silver",
    "the color bronze",
    "the color amber",
    "the color burgundy",
    "the color chartreuse",
    "the color crimson",
    "the color cyan",
    "the color fuchsia",
    "the color lavender",
    "the color lime",
    "the color mint",
    "the color navy",
    "the color olive",
    "the color peach",
    "the color plum",
    "the color salmon",
    "the color sienna",
    "the color tan",
    "the color taupe",
    "the color periwinkle",
    "the color aquamarine",
    "the color beige",
    "the color coral",
    "the color eggplant",
    "the color ivory",
    "the color khaki",
    "the color lemon",
    "the color lilac",
    "the color mocha",
    "the color mustard",
    "the color ruby",
    "the color sapphire",
    # Group 2: Animals (50)
    "cats",
    "dogs",
    "elephants",
    "lions",
    "tigers",
    "bears",
    "wolves",
    "foxes",
    "horses",
    "cows",
    "pigs",
    "sheep",
    "goats",
    "chickens",
    "ducks",
    "geese",
    "deer",
    "monkeys",
    "giraffes",
    "zebras",
    "kangaroos",
    "pandas",
    "rabbits",
    "squirrels",
    "bats",
    "whales",
    "dolphins",
    "sharks",
    "octopuses",
    "crabs",
    "lobsters",
    "snakes",
    "alligators",
    "crocodiles",
    "parrots",
    "eagles",
    "falcons",
    "peacocks",
    "ostriches",
    "camels",
    "buffaloes",
    "rhinos",
    "hippopotamuses",
    "ants",
    "bees",
    "butterflies",
    "spiders",
    "scorpions",
    "turtles",
    "penguins",
    # Group 3: Sports (50)
    "playing soccer",
    "playing basketball",
    "playing baseball",
    "playing football",
    "playing tennis",
    "playing volleyball",
    "playing golf",
    "playing hockey",
    "playing cricket",
    "playing rugby",
    "playing table tennis",
    "playing badminton",
    "playing squash",
    "playing lacrosse",
    "skiing",
    "snowboarding",
    "surfing",
    "skateboarding",
    "cycling",
    "running",
    "swimming",
    "boxing",
    "wrestling",
    "fencing",
    "martial arts",
    "archery",
    "climbing",
    "hiking",
    "kayaking",
    "canoeing",
    "sailing",
    "rowing",
    "diving",
    "scuba diving",
    "ice skating",
    "roller skating",
    "mountain biking",
    "trail running",
    "triathlon",
    "marathon running",
    "parkour",
    "ultimate frisbee",
    "playing handball",
    "playing darts",
    "playing billiards",
    "playing bowling",
    "playing ping pong",
    "playing ice hockey",
    "playing water polo",
    "playing field hockey",
    # Group 4: Household Activities/Objects (50)
    "washing the dishes",
    "cleaning the windows",
    "vacuuming the floor",
    "mopping the floor",
    "ironing clothes",
    "making the bed",
    "doing the laundry",
    "dusting the furniture",
    "taking out the trash",
    "cooking dinner",
    "baking bread",
    "washing the car",
    "gardening",
    "raking the leaves",
    "watering the plants",
    "pruning the bushes",
    "sweeping the porch",
    "polishing the silver",
    "organizing the closet",
    "cleaning the bathroom",
    "scrubbing the tub",
    "cleaning the refrigerator",
    "emptying the dishwasher",
    "making coffee",
    "setting the table",
    "preparing breakfast",
    "rearranging the furniture",
    "ironing the curtains",
    "folding the laundry",
    "mending clothes",
    "recycling",
    "composting",
    "decluttering",
    "cleaning the gutters",
    "cleaning the patio",
    "cleaning the oven",
    "organizing the pantry",
    "cleaning the carpets",
    "dusting the shelves",
    "cleaning the ceiling fan",
    "cleaning the microwave",
    "scrubbing the floor",
    "polishing the wood",
    "cleaning the blinds",
    "watering the garden",
    "weeding the garden",
    "cleaning the garage",
    "sweeping the driveway",
    "rinsing the vegetables",
    "chopping vegetables",
    # Group 5: Concepts (50)
    "philosophy",
    "mathematics",
    "history",
    "science",
    "literature",
    "poetry",
    "art",
    "music",
    "cinema",
    "theatre",
    "technology",
    "astronomy",
    "psychology",
    "sociology",
    "economics",
    "politics",
    "ecology",
    "anthropology",
    "ethics",
    "religion",
    "logic",
    "metaphysics",
    "biology",
    "geography",
    "linguistics",
    "archaeology",
    "literacy",
    "creativity",
    "imagination",
    "innovation",
    "curiosity",
    "adventure",
    "mystery",
    "beauty",
    "courage",
    "compassion",
    "empathy",
    "integrity",
    "kindness",
    "patience",
    "wisdom",
    "resilience",
    "determination",
    "gratitude",
    "humor",
    "optimism",
    "simplicity",
    "elegance",
    "serenity",
    "balance",
    # Group 6: Music/Arts (50)
    "playing guitar",
    "playing piano",
    "playing drums",
    "playing violin",
    "playing flute",
    "playing saxophone",
    "singing",
    "dancing",
    "painting",
    "drawing",
    "sculpting",
    "playing the trumpet",
    "playing the clarinet",
    "playing the cello",
    "playing the harp",
    "playing the oboe",
    "playing the accordion",
    "playing the banjo",
    "playing the mandolin",
    "playing the ukulele",
    "composing music",
    "writing poetry",
    "writing stories",
    "writing novels",
    "performing stand-up comedy",
    "photography",
    "filmmaking",
    "acting",
    "ballet dancing",
    "hip-hop dancing",
    "modern dancing",
    "tap dancing",
    "jazz dancing",
    "playing percussion",
    "singing opera",
    "DJing",
    "rapping",
    "beatboxing",
    "mixing music",
    "recording music",
    "listening to classical music",
    "listening to rock music",
    "listening to jazz",
    "listening to blues",
    "listening to electronic music",
    "listening to folk music",
    "performing magic",
    "calligraphy",
    "sketching",
    "playing board games",
    # Group 7: Vehicles/Travel (50)
    "cars",
    "bicycles",
    "motorcycles",
    "trains",
    "airplanes",
    "boats",
    "scooters",
    "submarines",
    "helicopters",
    "trucks",
    "vans",
    "sailboats",
    "kayaks",
    "canoes",
    "jet skis",
    "electric scooters",
    "skateboards",
    "rollerblades",
    "segways",
    "hot air balloons",
    "space shuttles",
    "ferries",
    "limousines",
    "convertibles",
    "minivans",
    "motorhomes",
    "tractors",
    "bulldozers",
    "forklifts",
    "subways",
    "trams",
    "pedicabs",
    "rickshaws",
    "sailplanes",
    "gliders",
    "balloons",
    "camping trailers",
    "snowmobiles",
    "all-terrain vehicles",
    "roadsters",
    "coupe cars",
    "convertible cars",
    "electric cars",
    "hybrid cars",
    "sports cars",
    "off-road vehicles",
    "luxury sedans",
    "minibikes",
    "electric bikes",
    "racing cars",
    # Group 8: Food and Beverages (50)
    "pizza",
    "sushi",
    "pasta",
    "burger",
    "salad",
    "steak",
    "chocolate",
    "ice cream",
    "coffee",
    "tea",
    "wine",
    "beer",
    "cocktails",
    "bread",
    "cheese",
    "fruit salad",
    "vegetable soup",
    "sandwiches",
    "tacos",
    "burritos",
    "dumplings",
    "noodles",
    "curry",
    "barbecue",
    "hot dogs",
    "soup",
    "pastries",
    "cupcakes",
    "cookies",
    "brownies",
    "muffins",
    "pancakes",
    "waffles",
    "omelettes",
    "smoothies",
    "salmon",
    "grilled chicken",
    "roast beef",
    "french fries",
    "popcorn",
    "nachos",
    "cheesecake",
    "pie",
    "candy",
    "sandwich",
    "stir-fry",
    "lasagna",
    "quiche",
    "salsa",
    "guacamole",
    # Group 9: Hobbies/Leisure (50)
    "reading mystery novels",
    "reading science fiction",
    "reading fantasy novels",
    "solving puzzles",
    "playing chess",
    "playing checkers",
    "playing video games",
    "building model airplanes",
    "building model cars",
    "collecting stamps",
    "collecting coins",
    "collecting postcards",
    "collecting antiques",
    "writing journals",
    "blogging",
    "vlogging",
    "scrapbooking",
    "knitting",
    "crocheting",
    "sewing",
    "embroidery",
    "painting miniatures",
    "solving crossword puzzles",
    "playing sudoku",
    "bird watching",
    "stargazing",
    "astronomy hobby",
    "fishing",
    "camping",
    "woodworking",
    "metalworking",
    "pottery making",
    "baking cakes",
    "brewing beer",
    "winemaking",
    "cheese making",
    "travel blogging",
    "geocaching",
    "roller coaster riding",
    "visiting museums",
    "visiting art galleries",
    "visiting historical sites",
    "learning languages",
    "dancing salsa",
    "dancing tango",
    "learning magic tricks",
    "solving riddles",
    "visiting theme parks",
    "exploring caves",
    "attending concerts",
    # Group 10: Miscellaneous (50)
    "modern architecture",
    "robotics",
    "quantum physics",
    "meditation",
    "yoga",
    "traveling",
    "cooking experiments",
    "urban exploration",
    "digital art",
    "virtual reality",
    "augmented reality",
    "3D printing",
    "astronautics",
    "cybersecurity",
    "artificial intelligence",
    "machine learning",
    "data science",
    "environmental conservation",
    "sustainable living",
    "minimalism",
    "vintage cars",
    "classic literature",
    "indie films",
    "experimental theater",
    "improv theatre",
    "modern dance",
    "street art",
    "graffiti art",
    "podcasting",
    "social media trends",
    "entrepreneurship",
    "investment strategies",
    "stock market trends",
    "cryptocurrency",
    "travel photography",
    "wildlife conservation",
    "botany",
    "zoology",
    "geology",
    "mythology",
    "folklore",
    "origami",
    "historical documentaries",
    "space exploration",
    "bird photography",
    "landscape photography",
    "comic books",
    "anime",
    "manga",
    "virtual concerts",
]



from dataclasses import dataclass
from random import Random
from typing import Any, Optional

from ..coaching import BaseCurriculum, ScalarAttributeDefinition
from ..factory import ProceduralDataset, register_dataset
from .contrib.logic_puzzle.generate import generate_puzzle

DATASET_NAME = "zebra_puzzles"


@dataclass
class ZebraConfig:
    """Configuration for zebra puzzle generation"""

    num_people: int = 4
    num_characteristics: int = 4
    seed: Optional[int] = None
    size: int = 500

    def validate(self):
        """Validate configuration parameters"""
        assert 2 <= self.num_people <= 7, "num_people must be between 2 and 7"
        assert 2 <= self.num_characteristics <= 7, "num_characteristics must be between 2 and 7"


class ZebraDataset(ProceduralDataset):
    """Generates [Zebra Puzzles](https://en.wikipedia.org/wiki/Zebra_Puzzle) with configurable parameters"""

    def __init__(self, config: ZebraConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)

    def __getitem__(self, idx: int) -> dict:
        """Generate a single Zebra task

        Returns:
            dict with keys:
                - question: str, the task description
                - answer: str, a solution string
                - metadata: dict with generation parameters
        """
        rng = Random(self.seed + idx)

        K = self.config.num_people
        M = self.config.num_characteristics
        instance, puzzle = generate_puzzle(rng, K, M)
        q = instance["questions"][0]["question"]
        answer = instance["questions"][0]["answer"]
        question = str(puzzle) + "\n" + q
        question = question + "? Provide only the name of the person as your final answer."

        return {
            "question": question,
            "answer": answer,
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "difficulty": {"num_people": K, "num_characteristics": M},
            },
        }

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        """Determine if the solution provided solves the Zebra task.

        The function awards 1.0 for a correct answer.

        Args:
            answer (Optional[str]): The user's answer.
            entry (dict[str, Any]): The original dataset entry containing the correct answer.

        Returns:
            float: The computed score between 0.0 and 1.0.
        """

        if isinstance(answer, str):
            if answer.lower().replace("\n", "") == entry["answer"].lower().replace("\n", ""):
                return 1.0  # Yay
        return 0.0


class ZebraCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(ZebraCurriculum.__name__, ZebraConfig)
        self._define_attributes(
            ScalarAttributeDefinition(
                name="num_people",
                levels=list(range(2, 8)),
                description="The number of people in the Zebra puzzle",
                field_name="num_people",
            ),
            ScalarAttributeDefinition(
                name="num_characteristics",
                levels=list(range(2, 8)),
                description="The number of characteristics in the Zebra puzzle",
                field_name="num_characteristics",
            ),
        )


register_dataset(DATASET_NAME, ZebraDataset, ZebraConfig, ZebraCurriculum)



from heapq import heappop, heappush

import numpy as np


def print_state(state, shape):
    if not state:
        return
    m, n = shape
    matrix = np.array(list(state)).reshape(m, n)
    print(matrix)


def find_boxes_and_goals(state, shape):
    _, width = shape
    boxes, goals, boxes_on_goal = [], [], []
    for pos, char in enumerate(state):
        if char == "@":
            boxes.append((pos // width, pos % width))
        elif char in "X%":
            goals.append((pos // width, pos % width))
        elif char == "$":
            boxes_on_goal.append((pos // width, pos % width))
    return boxes, goals, boxes_on_goal


def get_state(matrix):
    return matrix.tobytes().decode("utf-8").replace("\x00", "")


def is_solved(state):
    return "@" not in state


def manhattan_sum(state, player_pos, shape):
    height, width = shape
    player_x, player_y = player_pos
    boxes, goals, _ = find_boxes_and_goals(state, shape)
    boxes_cost = len(boxes) * height * width
    player_cost = 0
    for box_x, box_y in boxes:
        boxes_cost += min(abs(box_x - goal_x) + abs(box_y - goal_y) for goal_x, goal_y in goals)
    player_cost = min(abs(box_x - player_x) + abs(box_y - player_y) for box_x, box_y in boxes) if boxes else 0
    return boxes_cost + player_cost


def dijkstra(state, shape, box_pos=None, player_pos=None):
    height, width = shape
    dijk = np.array([[float("inf") for _ in range(width)] for _ in range(height)])
    dijk[box_pos or player_pos] = 0
    moves = [(1, 0), (-1, 0), (0, 1), (0, -1)]
    heap = [(0, box_pos or player_pos)]
    obstacles = "+" if player_pos else "+@$"
    while heap:
        distance, curr_pos = heappop(heap)
        if distance > dijk[curr_pos]:
            continue
        for move in moves:
            new_x, new_y = curr_pos[0] + move[0], curr_pos[1] + move[1]
            new_pos = new_x, new_y
            if 1 <= new_x < height - 1 and 1 <= new_y < width - 1 and state[new_x * width + new_y] not in obstacles:
                new_distance = distance + 1
                if new_distance < dijk[new_pos]:
                    dijk[new_pos] = new_distance
                    heappush(heap, (new_distance, new_pos))
    return dijk


def dijkstra_sum(state, player_pos, shape, distances):
    height, width = shape
    boxes, goals, boxes_on_goal = find_boxes_and_goals(state, shape)
    boxes_cost = len(boxes) * height * width
    player_cost = 0
    for box in boxes + boxes_on_goal:
        distances[box] = dijkstra(state, shape, box)
    distances[player_pos] = dijkstra(state, shape, player_pos=player_pos)
    for box in boxes:
        boxes_cost += min(distances[box][goal] for goal in goals)
    player_cost = min(distances[player_pos][box] for box in boxes) if boxes else 0
    return boxes_cost + player_cost


def is_deadlock(state, shape):
    height, width = shape
    if not state or len(state) != height * width:
        return False
    boxes, _, _ = find_boxes_and_goals(state, shape)
    for bx, by in boxes:  # corner deadlock
        box = bx * width + by
        if (
            (state[box - 1] == "+" and state[box - width] == "+")
            or (state[box + 1] == "+" and state[box + width] == "+")
            or (state[box + 1] == "+" and state[box - width] == "+")
            or (state[box - 1] == "+" and state[box + width] == "+")
        ):
            return True
    double_box_positions = [
        (0, -1, -width, -width - 1),
        (0, 1, -width, -width + 1),
        (0, -1, width - 1, width),
        (0, 1, width + 1, width),
    ]
    for bx, by in boxes:  # double box deadlock
        box = bx * width + by
        for pos in double_box_positions:
            pos_set = set()
            for dir in pos:
                pos_set.add(state[box + dir])
            if pos_set in ({"@", "+"}, {"@"}, {"@", "$"}, {"@", "$", "+"}):
                return True
    box = goal = 0
    for i in range(width + 1, 2 * width - 1):  # too many boxes deadlock
        if state[i] == "@":
            box += 1
        elif state[i] in "X%":
            goal += 1
    if box > goal:
        return True
    box = goal = 0
    for i in range(width * (height - 2) + 1, width * (height - 2) + width - 1):
        if state[i] == "@":
            box += 1
        elif state[i] in "X%":
            goal += 1
    if box > goal:
        return True
    box = goal = 0
    for i in range(width + 1, width * (height - 1) + 1, width):
        if state[i] == "@":
            box += 1
        elif state[i] in "X%":
            goal += 1
    if box > goal:
        return True
    box = goal = 0
    for i in range(2 * width - 2, width * height - 2, width):
        if state[i] == "@":
            box += 1
        elif state[i] in "X%":
            goal += 1
    if box > goal:
        return True
    return False


def can_move(state, shape, player_pos, move):
    new_state = list(state)
    x, y = player_pos
    _, width = shape
    move_cost = 0
    target = x + move[0], y + move[1]
    boxtarget = x + move[0] * 2, y + move[1] * 2
    curr1d = x * width + y
    target1d = target[0] * width + target[1]
    boxtarget1d = boxtarget[0] * width + boxtarget[1]
    if state[target1d] == "+":
        return None, move_cost
    elif state[target1d] in "-X":
        new_state[curr1d] = "-" if new_state[curr1d] == "*" else "X"
        new_state[target1d] = "*" if new_state[target1d] == "-" else "%"
        move_cost = 3
    elif state[target1d] in "@$":
        if state[boxtarget1d] in "+@$":
            return None, move_cost
        elif state[boxtarget1d] in "-X":
            new_state[boxtarget1d] = "@" if new_state[boxtarget1d] == "-" else "$"
            new_state[target1d] = "*" if new_state[target1d] == "@" else "%"
            new_state[curr1d] = "-" if new_state[curr1d] == "*" else "X"
            move_cost = 0 if new_state[boxtarget1d] == "$" else 2
    return "".join(new_state), move_cost



"""Count number of 1 bits in a number."""

from dataclasses import dataclass
from random import Random
from typing import Optional

from ..coaching import BaseCurriculum, RangeAttributeDefinition
from ..factory import ProceduralDataset, register_dataset

QUESTION_TEMPLATE = """How many 1 bits are there in the binary representation of the number {number}?"""

DATASET_NAME = "count_bits"


@dataclass
class CountBitsConfig:
    """Configuration for Count Bits dataset generation"""

    min_n: int = 1  # Minimum number to consider
    max_n: int = 2**31 - 1  # Maximum number to consider

    size: int = 500  # Virtual dataset size
    seed: Optional[int] = None

    def validate(self):
        """Validate configuration parameters"""
        assert 1 <= self.min_n <= self.max_n, "min_n must be between 1 and max_n"


class CountBitsDataset(ProceduralDataset):
    """Generates Count Bits exercises with configurable difficulty"""

    def __init__(self, config: CountBitsConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)

    def __getitem__(self, idx: int) -> dict:
        """Generate a single Count Bits question"""
        rng = Random(self.seed + idx)

        number = rng.randint(self.config.min_n, self.config.max_n)
        binary = bin(number)[2:]
        answer = binary.count("1")

        return {
            "question": QUESTION_TEMPLATE.format(number=number),
            "answer": str(answer),
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "number": number,
                "solution": answer,
                "binary": binary,
                "n": number,
                "difficulty": {
                    "n": (self.config.min_n, self.config.max_n),
                },
            },
        }


class CountBitsCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(CountBitsCurriculum.__name__, CountBitsConfig)

        # Define attributes
        self._define_attributes(
            RangeAttributeDefinition(
                name="n",
                levels=[10, 1_000, 1_000_000, 100_000_000, 2**31 - 1],
                description="Number to count bits in",
                lower_field_name="min_n",
                upper_field_name="max_n",
                ensure_interval=True,
            ),
        )


register_dataset(DATASET_NAME, CountBitsDataset, CountBitsConfig, CountBitsCurriculum)



import abc
from collections.abc import Iterable
from typing import Any, Optional, TypeVar

from ..utils import StrEnum
from .attributes import AttributeDefinition, RangeAttributeDefinition, ScalarAttributeDefinition

ConfigT = TypeVar("ConfigT")


class CurriculumContext(abc.ABC):
    @abc.abstractmethod
    def get_attr_value(self, curriculum, attr: AttributeDefinition) -> Any:
        pass


class RangeAttributeMode(StrEnum):
    """Text transformation options"""

    UPPER_BOUND = "upper_bound"  # only use the highest range segment
    INCLUSIVE = "inclusive"  # include all previous levels
    LAST_K = "last_k"  # use only the last *k* difficulty levels


class DefaultCurriculumContext(CurriculumContext):
    def __init__(self, mode: RangeAttributeMode = RangeAttributeMode.INCLUSIVE, k: int | None = None):
        """Create a CurriculumContext used when turning a curriculum into a concrete configuration.

        Args:
            mode: Strategy for translating a RangeAttributeDefinition level into a concrete range.
            k:   When ``mode`` is ``RangeAttributeMode.LAST_K`` this parameter indicates how many of the
                 most-recent difficulty levels (counting backwards from the current one) should be kept.
                 If ``k`` is ``None`` the behaviour falls back to ``INCLUSIVE`` (i.e. keep everything).  The
                 parameter is ignored for other modes.
        """
        self.mode = mode
        self.k = k  # window size used for LAST_K mode

    def get_range_attr_value(self, curriculum, attr: RangeAttributeDefinition) -> Any:
        level = curriculum.get_attr_level(attr.name)
        v = attr.get_level_value(level)
        if isinstance(v, Iterable):
            return v

        if attr.ensure_interval:
            if self.mode == RangeAttributeMode.UPPER_BOUND:
                hi_index = min(level + 1, len(attr.levels) - 1)
                lo_index = max(0, hi_index - 1)

            elif self.mode == RangeAttributeMode.INCLUSIVE:
                lo_index = 0
                hi_index = min(level + 1, len(attr.levels) - 1)

            elif self.mode == RangeAttributeMode.LAST_K:
                hi_index = min(level, len(attr.levels) - 1)
                window = self.k if self.k is not None else hi_index + 1
                lo_index = max(0, hi_index - window + 1)

        else:
            if self.mode == RangeAttributeMode.UPPER_BOUND:
                hi_index = min(level, len(attr.levels) - 1)
                lo_index = max(0, hi_index)

            elif self.mode == RangeAttributeMode.INCLUSIVE:
                lo_index = 0
                hi_index = min(level, len(attr.levels) - 1)

            elif self.mode == RangeAttributeMode.LAST_K:
                hi_index = min(level, len(attr.levels) - 1)
                window = self.k if self.k is not None else hi_index + 1
                lo_index = max(0, hi_index - window + 1)

        # Additional handling for LAST_K when ensure_interval is True (above branch)
        if attr.ensure_interval and self.mode == RangeAttributeMode.LAST_K:
            # Re-compute indices so that we always return at least a two-value interval
            if hi_index == lo_index:
                lo_index = max(0, hi_index - 1)

        lo = attr.get_level_value(lo_index)
        hi = attr.get_level_value(hi_index)

        return [lo, hi]

    def get_attr_value(self, curriculum, attr: AttributeDefinition) -> Any:
        if isinstance(attr, RangeAttributeDefinition):
            return self.get_range_attr_value(curriculum, attr)
        elif isinstance(attr, ScalarAttributeDefinition):
            return curriculum.get_attr_value(attr.name)


class BaseCurriculum:
    def __init__(self, name: str, config_cls: ConfigT):
        self.name = name
        self._config_cls = config_cls
        self._attributes: dict[str, AttributeDefinition] = {}
        self._current_levels: dict[str, int] = {}

    def generate_configuration(
        self, defaults: Optional[dict[str, Any]] = None, context: Optional[CurriculumContext] = None
    ) -> ConfigT:
        config_args = defaults.copy() if defaults is not None else {}

        if context is None:
            context = DefaultCurriculumContext(mode=RangeAttributeMode.INCLUSIVE)

        for attr in self._attributes.values():
            if isinstance(attr, RangeAttributeDefinition):
                v = context.get_attr_value(self, attr)
                if not isinstance(v, Iterable):
                    v = [v]
                config_args[attr.lower_field_name] = min(v)
                config_args[attr.upper_field_name] = max(v)
            elif isinstance(attr, ScalarAttributeDefinition):
                val = context.get_attr_value(self, attr)
                config_args[attr.field_name] = val
        return self._config_cls(**config_args)

    @property
    def attributes(self) -> dict[str, AttributeDefinition]:
        """Get the curriculum's attributes"""
        return self._attributes

    def get_attribute(self, attr_name: str) -> AttributeDefinition:
        if attr_name not in self._attributes:
            raise KeyError(f"Attribute '{self.name}.{attr_name}' does not exist")
        return self._attributes[attr_name]

    def _define_attributes(self, *attrs: tuple[AttributeDefinition, ...]) -> None:
        for attr in attrs:
            if attr.name in self.attributes:
                raise RuntimeError(f"Attribute with name {attr.name} is already defined.")
            self.attributes[attr.name] = attr

    def get_attr_level(self, attr_name: str) -> int:
        """
        Get the current level for an attribute.
        Args:
            attr_name: Name of the attribute
        Returns:
            Current level index for the attribute
        """
        attr = self.get_attribute(attr_name)
        return self._current_levels.get(attr_name, attr.default_level)

    def get_attr_value(self, attr_name: str) -> Any:
        """
        Get the current value for an attribute based on its level.
        Args:
            attr_name: Name of the attribute
        Returns:
            Current value for the attribute based on its level and type
        """
        attr = self.get_attribute(attr_name)
        level = self.get_attr_level(attr_name)
        return attr.get_level_value(level)

    def set_attr_level(self, attr_name: str, level: int) -> None:
        """
        Set the level for an attribute.
        Args:
            attr_name: Name of the attribute
            level: New level index
        """
        attr = self.get_attribute(attr_name)
        attr.validate_level(level, curriculum=self.name)
        self._current_levels[attr_name] = level

    def increment_attr_level(self, attr_name: str) -> bool:
        """
        Increment the level of an attribute if possible.
        Args:
            attr_name: Name of the attribute to increment
        Returns:
            bool: True if level was incremented, False if already at max level
        Raises:
            KeyError: If attribute doesn't exist
        """
        attr = self.get_attribute(attr_name)
        current_level = self.get_attr_level(attr_name)
        if current_level < len(attr.levels) - 1:
            self.set_attr_level(attr_name, current_level + 1)
            return True
        return False

    def decrement_attr_level(self, attr_name: str) -> bool:
        """
        Decrement the level of an attribute if possible.
        Args:
            attr_name: Name of the attribute to decrement
        Returns:
            bool: True if level was decremented, False if already at min level
        Raises:
            KeyError: If attribute doesn't exist
        """
        current_level = self.get_attr_level(attr_name)
        if current_level > 0:
            self.set_attr_level(attr_name, current_level - 1)
            return True
        return False

    def get_max_level(self) -> int:
        """
        Get the maximum level currently set across all attributes.

        Returns:
            int: The maximum level currently set across all attributes
        """
        if not self._attributes:
            return 0

        return max(self.get_attr_level(attr_name) for attr_name in self._attributes)

    def set_global_level(self, level: int) -> None:
        """
        Set all attributes to the specified level.
        If the level exceeds the number of defined levels for an attribute,
        use the highest defined level for that attribute.

        Args:
            level: The level to set for all attributes
        """
        for attr_name, attr in self._attributes.items():
            # Use the highest defined level if the requested level exceeds available levels
            attr_level = min(level, len(attr.levels) - 1)
            self.set_attr_level(attr_name, attr_level)

    def increment_global_level(self) -> bool:
        """
        Increment the level of all attributes by one from the current maximum level.

        Returns:
            bool: True if at least one attribute's level was incremented, False otherwise
        """
        current_max = self.get_max_level()
        target_level = current_max + 1

        # Check if any attribute can be incremented
        can_increment = any(
            self.get_attr_level(attr_name) < len(self._attributes[attr_name].levels) - 1
            for attr_name in self._attributes
        )

        if can_increment:
            for attr_name, attr in self._attributes.items():
                # Only increment if the attribute is not already at its maximum level
                if self.get_attr_level(attr_name) < len(attr.levels) - 1:
                    # Don't exceed the attribute's maximum level
                    new_level = min(target_level, len(attr.levels) - 1)
                    self.set_attr_level(attr_name, new_level)
            return True
        return False

    def decrement_global_level(self) -> bool:
        """
        Decrement the level of all attributes by one from the current maximum level.

        Returns:
            bool: True if at least one attribute's level was decremented, False otherwise
        """
        current_max = self.get_max_level()

        if current_max > 0:
            target_level = current_max - 1
            for attr_name in self._attributes:
                # Only decrement if the attribute is at the current maximum level
                if self.get_attr_level(attr_name) == current_max:
                    self.set_attr_level(attr_name, target_level)
            return True
        return False

    def get_global_level(self) -> Optional[int]:
        """Get the global level of the curriculum."""
        attr_dict = {}
        if not self._attributes:
            return 0
        for attr_name in self._attributes:
            attr = self.get_attribute(attr_name)
            if isinstance(attr, RangeAttributeDefinition):
                attr_dict[attr.upper_field_name] = self.get_attr_value(attr_name)
            elif isinstance(attr, ScalarAttributeDefinition):
                attr_dict[attr.field_name] = self.get_attr_value(attr_name)
        return attr_dict



import random
from dataclasses import dataclass
from typing import Any, Optional

from ..coaching import BaseCurriculum, RangeAttributeDefinition
from ..factory import ProceduralDataset, register_dataset
from ..utils import StrEnum


class Color(StrEnum):
    RED = "red"
    GREEN = "green"
    BLUE = "blue"
    YELLOW = "yellow"
    WHITE = "white"
    ORANGE = "orange"
    PURPLE = "purple"
    PINK = "pink"
    BROWN = "brown"
    GRAY = "gray"
    CYAN = "cyan"
    MAGENTA = "magenta"
    GOLD = "gold"
    SILVER = "silver"
    INDIGO = "indigo"
    VIOLET = "violet"


class Side(StrEnum):
    TOP = "top"
    RIGHT = "right"
    FRONT = "front"
    LEFT = "left"
    BACK = "back"
    BOTTOM = "bottom"


DATASET_NAME = "color_cube_rotation"


@dataclass
class Cube:
    """Represents a cube with colored sides"""

    colors: dict[Side, Color]

    def rotate_front_to_top(self) -> None:
        """Rotate cube so front face becomes top"""
        old = self.colors.copy()
        self.colors[Side.TOP] = old[Side.FRONT]
        self.colors[Side.FRONT] = old[Side.BOTTOM]
        self.colors[Side.BOTTOM] = old[Side.BACK]
        self.colors[Side.BACK] = old[Side.TOP]
        # Right and left stay in place

    def rotate_right_to_top(self) -> None:
        """Rotate cube so right face becomes top"""
        old = self.colors.copy()
        self.colors[Side.TOP] = old[Side.RIGHT]
        self.colors[Side.RIGHT] = old[Side.BOTTOM]
        self.colors[Side.BOTTOM] = old[Side.LEFT]
        self.colors[Side.LEFT] = old[Side.TOP]
        # Front and back stay in place

    def rotate_back_to_top(self) -> None:
        """Rotate cube so back face becomes top"""
        old = self.colors.copy()
        self.colors[Side.TOP] = old[Side.BACK]
        self.colors[Side.BACK] = old[Side.BOTTOM]
        self.colors[Side.BOTTOM] = old[Side.FRONT]
        self.colors[Side.FRONT] = old[Side.TOP]
        # Right and left stay in place

    def rotate_left_to_top(self) -> None:
        """Rotate cube so left face becomes top"""
        old = self.colors.copy()
        self.colors[Side.TOP] = old[Side.LEFT]
        self.colors[Side.LEFT] = old[Side.BOTTOM]
        self.colors[Side.BOTTOM] = old[Side.RIGHT]
        self.colors[Side.RIGHT] = old[Side.TOP]
        # Front and back stay in place

    def rotate_bottom_to_top(self) -> None:
        """Rotate cube so bottom face becomes top"""
        old = self.colors.copy()
        self.colors[Side.TOP] = old[Side.BOTTOM]
        self.colors[Side.BOTTOM] = old[Side.TOP]
        self.colors[Side.FRONT] = old[Side.BACK]
        self.colors[Side.BACK] = old[Side.FRONT]
        # Right and left stay in place


@dataclass
class ColorCubeRotationConfig:
    """Configuration for color cube rotation task generation"""

    min_rotations: int = 1
    max_rotations: int = 3
    seed: Optional[int] = None
    size: int = 500

    def validate(self) -> None:
        """Validate configuration parameters"""
        assert self.min_rotations > 0, "min_rotations must be positive"
        assert self.max_rotations >= self.min_rotations, "max_rotations must be >= min_rotations"


class ColorCubeRotationDataset(ProceduralDataset):
    """Generates color cube rotation reasoning tasks"""

    def __init__(self, config: ColorCubeRotationConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)

    def __getitem__(self, idx: int) -> dict:
        rng = random.Random(self.seed + idx)

        # Generate initial cube state
        cube = self._generate_cube(rng)
        initial_state = cube.colors.copy()

        # Generate sequence of rotations
        num_rotations = rng.randint(self.config.min_rotations, self.config.max_rotations)
        rotations = []

        # Keep trying until we have at least one valid rotation
        while len(rotations) < num_rotations:
            # Get all sides except TOP
            available_sides = [s for s in Side if s != Side.TOP]
            from_side = rng.choice(available_sides)
            rotations.append(from_side)
            self._rotate_to_top(cube, from_side)

        # Select target side for question
        target_side = rng.choice(list(Side))

        # Generate story
        story = self._generate_story(initial_state, rotations, target_side, rng)

        return {
            "question": story,
            "answer": cube.colors[target_side].value,
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "initial_state": {k.value: v.value for k, v in initial_state.items()},
                "rotations": [r.value for r in rotations],
                "target_side": target_side.value,
                "num_rotations": num_rotations,
                "difficulty": {
                    "rotations": (self.config.min_rotations, self.config.max_rotations),
                },
            },
        }

    def _generate_cube(self, rng: random.Random) -> Cube:
        """Generate a cube with random colors"""
        colors = list(Color)
        rng.shuffle(colors)  # Randomize color order
        return Cube({side: color for side, color in zip(Side, colors)})

    def _rotate_to_top(self, cube: Cube, from_side: Side) -> None:
        """Rotate cube so that given side becomes top"""
        rotation_map = {
            Side.FRONT: cube.rotate_front_to_top,
            Side.RIGHT: cube.rotate_right_to_top,
            Side.BACK: cube.rotate_back_to_top,
            Side.LEFT: cube.rotate_left_to_top,
            Side.BOTTOM: cube.rotate_bottom_to_top,
        }
        if from_side in rotation_map:
            rotation_map[from_side]()

    def _generate_story(
        self, initial_state: dict[Side, Color], rotations: list[Side], target_side: Side, rng: random.Random
    ) -> str:
        """Generate story describing cube state and rotations"""
        # Describe initial state
        story_parts = ["A cube has:"]
        for side in Side:
            story_parts.append(f"- a {initial_state[side].value} {side.value} side")

        # Describe rotations
        rotation_templates = [
            "The cube is rotated so that the side which was before at the {side} is now at the top.",
            "Then the cube is rotated to bring the {side} side to the top.",
            "After that the cube is turned to make the {side} face the top.",
            "Now the cube is rotated to place its {side} side at the top.",
            "Next, the {side} side is rotated to become the top face.",
        ]

        for i, from_side in enumerate(rotations):
            template = rotation_templates[0] if i == 0 else rng.choice(rotation_templates[1:])
            story_parts.append(f"\n{template.format(side=from_side.value)}")

        # Ask question
        story_parts.append(f"\nWhat is now the color of the {target_side.value} side of the cube?")
        story_parts.append(f"Provide only the color as your final answer.")

        return "\n".join(story_parts)

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        reward = 0.0
        if answer is not None:
            try:
                answer_formatted = answer.strip().lower()
                solved = answer_formatted == entry["answer"].strip().lower()
                if solved:
                    reward = 1.0
                else:
                    reward = 0.01
            except:
                reward = 0.01
        return reward


class ColorCubeRotationCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(ColorCubeRotationCurriculum.__name__, ColorCubeRotationConfig)

        # Define attributes
        self._define_attributes(
            RangeAttributeDefinition(
                name="rotations",
                levels=[1, 5, 10, 50, 100],
                description="Number of rotations to perform on the cube",
                lower_field_name="min_rotations",
                upper_field_name="max_rotations",
                ensure_interval=True,
            )
        )


register_dataset(DATASET_NAME, ColorCubeRotationDataset, ColorCubeRotationConfig, ColorCubeRotationCurriculum)



"""Mini Sudoku (4x4) puzzle generator"""

import copy
from dataclasses import dataclass
from random import Random
from typing import Any, Optional

from ..coaching import BaseCurriculum, RangeAttributeDefinition
from ..factory import ProceduralDataset, register_dataset

DATASET_NAME = "mini_sudoku"


@dataclass
class MiniSudokuConfig:
    """Configuration for 4x4 sudoku puzzle generation"""

    min_empty: int = (
        8  # Minimum number of empty cells. Occasionally this can be violated, if removing more cells would break the puzzle's uniqueness.
    )
    max_empty: int = 12  # Maximum number of empty cells
    seed: Optional[int] = None
    size: int = 500  # Virtual dataset size

    def validate(self):
        """Validate configuration parameters"""
        # More than 12 empty cells is incompatible with a unique solution
        assert 0 <= self.min_empty <= 12, "min_empty must be between 0 and 12"
        assert self.min_empty <= self.max_empty <= 12, "max_empty must be between min_empty and 12"


class MiniSudokuDataset(ProceduralDataset):
    """Generates 4x4 sudoku puzzles with configurable difficulty"""

    def __init__(self, config: MiniSudokuConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)

    def __len__(self) -> int:
        return self.config.size

    def __iter__(self):
        self._current_idx = 0
        return self

    def __next__(self):
        if self._current_idx >= self.config.size:
            raise StopIteration
        item = self[self._current_idx]
        self._current_idx += 1
        return item

    def _is_valid(self, board: list[list[int]], row: int, col: int, num: int) -> bool:
        """Check if number can be placed at position"""
        # Check row
        if num in board[row]:
            return False

        # Check column
        if num in [board[i][col] for i in range(4)]:
            return False

        # Check 2x2 box
        box_row, box_col = 2 * (row // 2), 2 * (col // 2)
        for i in range(box_row, box_row + 2):
            for j in range(box_col, box_col + 2):
                if board[i][j] == num:
                    return False
        return True

    def _solve(self, board: list[list[int]]) -> bool:
        """Solve mini sudoku using backtracking"""
        empty = self._find_empty(board)
        if not empty:
            return True

        row, col = empty
        for num in range(1, 5):
            if self._is_valid(board, row, col, num):
                board[row][col] = num
                if self._solve(board):
                    return True
                board[row][col] = 0
        return False

    def _find_empty(self, board: list[list[int]]) -> Optional[tuple[int, int]]:
        """Find an empty cell"""
        for i in range(4):
            for j in range(4):
                if board[i][j] == 0:
                    return (i, j)
        return None

    def _generate_solved_board(self, rng: Random) -> list[list[int]]:
        """Generate a complete solved mini sudoku board"""
        board = [[0] * 4 for _ in range(4)]

        # Try multiple times to generate a valid board
        max_attempts = 100
        for _ in range(max_attempts):
            # Start fresh
            for i in range(4):
                for j in range(4):
                    board[i][j] = 0

            # Fill diagonal boxes first (they are independent)
            for i in range(0, 4, 2):
                nums = list(range(1, 5))
                rng.shuffle(nums)
                pos = 0
                for r in range(i, i + 2):
                    for c in range(i, i + 2):
                        board[r][c] = nums[pos]
                        pos += 1

            # Try to solve the rest
            if self._solve(board):
                return board

        raise RuntimeError("Failed to generate valid mini sudoku board")

    def _count_solutions(self, board: list[list[int]], limit: int = 2) -> int:
        """Count the number of solutions for a given board"""

        def _count_solutions_helper(board: list[list[int]]) -> int:
            empty = self._find_empty(board)
            if not empty:
                return 1

            row, col = empty
            count = 0
            for num in range(1, 5):
                if self._is_valid(board, row, col, num):
                    board[row][col] = num
                    count += _count_solutions_helper(board)
                    if count >= limit:
                        return count
                    board[row][col] = 0
            return count

        return _count_solutions_helper(board)

    def _create_puzzle(self, solved_board: list[list[int]], num_empty: int, rng: Random) -> list[list[int]]:
        """Create puzzle by removing numbers from solved board"""
        puzzle = [row[:] for row in solved_board]
        cells = [(i, j) for i in range(4) for j in range(4)]
        rng.shuffle(cells)
        num_removed = 0

        for i, j in cells:
            saved = puzzle[i][j]
            puzzle[i][j] = 0
            puzzle_copy = copy.deepcopy(puzzle)
            # Check if removing this clue breaks uniqueness
            if self._count_solutions(puzzle_copy) > 1:
                puzzle[i][j] = saved
            else:
                num_removed += 1
                if num_removed == num_empty:
                    break

        return puzzle

    def _board_to_string(self, board: list[list[int]]) -> str:
        """Convert board to string representation"""
        return "\n".join(" ".join(str(x) if x != 0 else "_" for x in row) for row in board)

    def __getitem__(self, idx: int) -> dict:
        """Generate a single mini sudoku puzzle"""
        rng = Random(self.seed + idx)

        # Generate solved board
        solved_board = self._generate_solved_board(rng)

        # Create puzzle by removing numbers
        num_empty = rng.randint(self.config.min_empty, self.config.max_empty)
        puzzle = self._create_puzzle(solved_board, num_empty, rng)

        # Update the num_empty to be used in the metadata if we couldn't remove as many as we wanted
        num_empty = sum(1 for row in puzzle for x in row if x == 0)

        # Format as strings
        puzzle_str = self._board_to_string(puzzle)
        solution_str = self._board_to_string(solved_board)

        question = (
            "In 4x4 Mini Sudoku:\n"
            "- Each row must contain each number from 1-4 exactly once\n"
            "- Each column must contain each number 1-4 exactly once\n"
            "- Each 2x2 subgrid must contain each number 1-4 exactly once\n"
            f"Solve this 4x4 Mini Sudoku puzzle:\n{puzzle_str}\n"
            "Format your response as the puzzle above, with spaces separating each number within a row, and newlines separating rows.\n"
        )

        return {
            "question": question,
            "answer": solution_str,
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "puzzle": puzzle,
                "solution": solved_board,
                "num_empty": num_empty,
                "difficulty": {
                    "empty": (self.config.min_empty, self.config.max_empty),
                },
            },
        }

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        if not isinstance(answer, str) or len(answer) == 0:
            return 0.0

        oracle_answer = entry["answer"]
        metadata = entry["metadata"]
        solution: list[list[int]] = metadata["solution"]
        board_size: int = len(solution[0])

        # 1. match answer without trailing whitespaces
        answer_stripped = "\n".join(l.rstrip() for l in answer.split("\n"))
        oracle_answer_stripped = "\n".join(l.rstrip() for l in oracle_answer.split("\n"))

        if answer_stripped == oracle_answer_stripped:
            reward = 1.0
        else:
            # 2. accept answers with correct numeric sequence (ignoring non-numeric characters)
            row = 0
            num_matching = 0
            for ln in answer.split("\n"):
                if row >= len(solution):
                    break
                numbers = [int(c) for c in ln if c in "123456789"]
                if len(numbers) != board_size:
                    continue  # ignore lines without numbers
                for a, b in zip(solution[row], numbers):
                    if a == b:
                        num_matching += 1
                row += 1

            reward = num_matching / (board_size * board_size)
            reward *= 0.9  # penalty for not using standard format

        if len(answer) > len(oracle_answer):
            reward *= len(oracle_answer) / len(answer)  # penalty for additional length
        return reward


class MiniSudokuCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(MiniSudokuCurriculum.__name__, MiniSudokuConfig)

        # Define attributes
        self._define_attributes(
            RangeAttributeDefinition(
                name="empty",
                levels=[4, 6, 8, 10, 12],
                description="Number of empty cells in the puzzle",
                lower_field_name="min_empty",
                upper_field_name="max_empty",
                ensure_interval=True,
            )
        )


register_dataset(DATASET_NAME, MiniSudokuDataset, MiniSudokuConfig, MiniSudokuCurriculum)



"""Go problem (tsumego) generator"""

"""
This module generates one-move Tsumego puzzles, which are Go problems focused on tactical capture scenarios.

The puzzles generated here have the following characteristics:
- They are created on a board of configurable size (with a minimum and maximum board size).
- A number of stones are randomly placed on the board, subject to a maximum stone limit.
- A specific capture problem is then constructed by arranging white stones in a plus-shaped formation.
- Extra liberties surrounding this white group are filled with black stones, except for one key liberty.
  This forces a situation where a single move by Black (at the remaining liberty) results in a capture.
- Puzzle generation is deterministic given a seed, which ensures reproducibility.

These puzzles are intended to provide focused practice on reading and executing capturing moves in Go.

TODO: Generate multi-step Tsumego problems.
"""

import re
from dataclasses import dataclass
from random import Random
from typing import Any, Optional

from ..coaching import BaseCurriculum, RangeAttributeDefinition, ScalarAttributeDefinition
from ..factory import ProceduralDataset, register_dataset

# Added constant to avoid repetition of adjacent directions
DIRECTIONS = [(-1, 0), (1, 0), (0, -1), (0, 1)]

DATASET_NAME = "tsumego"


@dataclass
class TsumegoConfig:
    """Configuration for Tsumego problem generation"""

    min_board_size: int = 9
    max_board_size: int = 13
    max_stones: int = 15
    size: int = 500
    seed: Optional[int] = None

    def __post_init__(self):
        """Validate configuration parameters"""
        if self.min_board_size < 5:
            raise ValueError("min_board_size must be at least 5")
        if self.max_board_size > 19:
            raise ValueError("max_board_size must be at most 19")
        if self.min_board_size > self.max_board_size:
            raise ValueError("min_board_size must be less than or equal to max_board_size")
        if self.max_stones < 5:
            raise ValueError("max_stones must be at least 5")


class TsumegoDataset(ProceduralDataset):
    """Generates Tsumego problems with configurable parameters"""

    def __init__(self, config: TsumegoConfig):
        self._prompt_templates = [
            "Tsumego time. Black to play and capture some stones.\nFind the key move.",
            "I have a Go problem for you. Black moves next - can you capture some of the white stones?",
            "Here's a Go challenge. Playing as Black, how can you capture as many white stones as possible?",
        ]
        self._ko_point = None
        super().__init__(config=config, seed=config.seed, size=config.size)

    # New helper method for board copying
    def _copy_board(self, board: list[list[str]]) -> list[list[str]]:
        """Return a deep copy of the board."""
        return [row[:] for row in board]

    def _get_liberties(self, board: list[list[str]], row: int, col: int) -> set[tuple[int, int]]:
        """Get empty adjacent points (liberties) for a stone"""
        size = len(board)
        liberties = set()
        for dr, dc in DIRECTIONS:
            r, c = row + dr, col + dc
            if 0 <= r < size and 0 <= c < size and board[r][c] == ".":
                liberties.add((r, c))
        return liberties

    def _get_group(self, board: list[list[str]], row: int, col: int) -> set[tuple[int, int]]:
        """Get all stones in the same group (connected stones of same color)"""
        size = len(board)
        color = board[row][col]
        if color == ".":
            return set()

        group = {(row, col)}
        queue = [(row, col)]
        while queue:
            r, c = queue.pop(0)
            for dr, dc in DIRECTIONS:
                nr, nc = r + dr, c + dc
                if 0 <= nr < size and 0 <= nc < size and board[nr][nc] == color and (nr, nc) not in group:
                    group.add((nr, nc))
                    queue.append((nr, nc))
        return group

    def _count_liberties(self, board: list[list[str]], group: set[tuple[int, int]]) -> int:
        """Count total liberties for a group of stones"""
        liberties = set()
        for row, col in group:
            liberties.update(self._get_liberties(board, row, col))
        return len(liberties)

    def _would_capture(self, board: list[list[str]], row: int, col: int, color: str) -> bool:
        """Check if a move would capture any opponent stones"""
        size = len(board)
        opponent = "O" if color == "X" else "X"

        # Make a copy of the board and place the stone
        board_copy = self._copy_board(board)
        board_copy[row][col] = color

        checked = set()
        for dr, dc in DIRECTIONS:
            r, c = row + dr, col + dc
            if 0 <= r < size and 0 <= c < size and board_copy[r][c] == opponent and (r, c) not in checked:
                group = self._get_group(board_copy, r, c)
                checked.update(group)
                if self._count_liberties(board_copy, group) == 0:
                    return True
        return False

    def _is_valid_move(self, board: list[list[str]], row: int, col: int, color: str) -> bool:
        """Check if a move is legal (not suicide, unless it captures)"""
        size = len(board)
        if not (0 <= row < size and 0 <= col < size):
            return False
        if board[row][col] != ".":
            return False
        if (row, col) == self._ko_point:
            return False

        # If the move captures opponent stones, it's valid
        if self._would_capture(board, row, col, color):
            return True

        board_copy = self._copy_board(board)
        board_copy[row][col] = color
        group = self._get_group(board_copy, row, col)
        return self._count_liberties(board_copy, group) > 0

    def _make_move(self, board: list[list[str]], row: int, col: int, color: str) -> bool:
        """Make a move and update ko point. Returns True if move was valid."""
        if not self._is_valid_move(board, row, col, color):
            return False

        self._ko_point = None
        board[row][col] = color
        opponent = "O" if color == "X" else "X"
        captured_stones = []

        for dr, dc in DIRECTIONS:
            r, c = row + dr, col + dc
            if 0 <= r < len(board) and 0 <= c < len(board) and board[r][c] == opponent:
                group = self._get_group(board, r, c)
                if self._count_liberties(board, group) == 0:
                    captured_stones.extend(group)

        if len(captured_stones) == 1 and len(self._get_group(board, row, col)) == 1:
            self._ko_point = captured_stones[0]

        for r, c in captured_stones:
            board[r][c] = "."

        return True

    def _generate_capture_problem(self, size: int, rng: Random) -> tuple[list[list[str]], tuple[int, int]]:
        """Generate a capture problem"""
        board = [["." for _ in range(size)] for _ in range(size)]
        stones_placed = 0
        max_stones = self.config.max_stones - 4  # Reserve space for capture setup

        while stones_placed < max_stones:
            row = rng.randint(0, size - 1)
            col = rng.randint(0, size - 1)
            color = "X" if rng.random() < 0.5 else "O"
            if board[row][col] == "." and self._is_valid_move(board, row, col, color):
                self._make_move(board, row, col, color)
                stones_placed += 1

        tries = 0
        formation_options = {
            "plus": {
                "white_offsets": [(0, 0), (-1, 0), (1, 0), (0, -1)],
                "forced_move_offset": (0, 1),
                "neighbor_offsets": [(0, 0), (-1, 0), (1, 0), (0, -1), (0, 1)],
            },
            "L": {
                "white_offsets": [(0, 0), (0, 1), (1, 0)],
                "forced_move_offset": (1, 1),
                "neighbor_offsets": [(0, 0), (0, 1), (1, 0), (1, 1)],
            },
            "T": {
                "white_offsets": [(0, -1), (0, 0), (0, 1), (1, 0)],
                "forced_move_offset": (-1, 0),
                "neighbor_offsets": [(0, -1), (0, 0), (0, 1), (1, 0), (-1, 0)],
            },
        }

        while tries < 50:
            row = rng.randint(1, size - 2)
            col = rng.randint(1, size - 2)
            formation_type = rng.choice(list(formation_options.keys()))
            formation = formation_options[formation_type]
            if all(board[row + dr][col + dc] == "." for dr, dc in formation["neighbor_offsets"]):
                # Place white stones according to chosen formation
                for dr, dc in formation["white_offsets"]:
                    board[row + dr][col + dc] = "O"
                forced_move = (row + formation["forced_move_offset"][0], col + formation["forced_move_offset"][1])
                white_group = {(row + dr, col + dc) for dr, dc in formation["white_offsets"]}
                extra_liberties = set()
                for r, c in white_group:
                    extra_liberties |= self._get_liberties(board, r, c)
                extra_liberties.discard(forced_move)
                for r, c in extra_liberties:
                    board[r][c] = "X"

                # Add decoy stone to enhance puzzle difficulty
                current_stone_count = sum(cell in "XO" for row in board for cell in row)
                if current_stone_count < self.config.max_stones + 7:
                    center = (row, col)  # using the base white stone as center
                    decoy_candidates = []
                    for i in range(center[0] - 2, center[0] + 3):
                        for j in range(center[1] - 2, center[1] + 3):
                            if abs(i - center[0]) + abs(j - center[1]) == 2:
                                if 0 <= i < size and 0 <= j < size and board[i][j] == "." and (i, j) != forced_move:
                                    decoy_candidates.append((i, j))
                    if decoy_candidates:
                        decoy_pos = rng.choice(decoy_candidates)
                        decoy_color = "X" if rng.random() < 0.5 else "O"
                        board[decoy_pos[0]][decoy_pos[1]] = decoy_color

                if self._is_valid_move(board, forced_move[0], forced_move[1], "X"):
                    return board, forced_move
            tries += 1
        raise RuntimeError("Failed to generate a capture problem")

    def _board_to_string(self, board: list[list[str]]) -> str:
        """Convert board to string representation"""
        size = len(board)
        # Column labels
        cols = "   " + " ".join(chr(ord("A") + i) for i in range(size)) + "\n"
        # Board with row numbers
        rows = [f"{size-i:2d} {' '.join(row)}" for i, row in enumerate(board)]
        return cols + "\n".join(rows)

    def __getitem__(self, idx: int) -> dict:
        """Generate a single Tsumego problem

        Returns:
            dict with:
            - "question": Problem description and board state
            - "answer": Solution move(s)
            - "metadata": Problem details and configuration
        """
        rng = Random(self.seed + idx if self.seed is not None else None)
        size = rng.randint(self.config.min_board_size, self.config.max_board_size)

        board, solution = self._generate_capture_problem(size, rng)
        board_str = self._board_to_string(board)
        solution_str = f"{chr(ord('A')+solution[1])}{size - solution[0]}"
        self._ko_point = None

        return {
            "question": (
                rng.choice(self._prompt_templates) + "\n\n```\n" + board_str + "\n```\n\n"
                "X - Black\n"
                "O - White\n\n"
                "Specify your move in coordinates (e.g. 'C4' for column C, row 4)"
            ),
            "answer": solution_str,
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "board": board,
                "board_size": size,
                "difficulty": {
                    "board_size": (self.config.min_board_size, self.config.max_board_size),
                },
            },
        }

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        """Score the answer against the solution"""

        oracle_answer = entry["answer"].strip()
        reward = 0.0
        if answer is not None and len(answer) > 0:
            answer = answer.strip().upper()
            if answer == oracle_answer:
                reward = 1.0
            elif oracle_answer in answer:
                reward = len(oracle_answer) / len(answer)
            elif re.match(r"^([A-Z])(\d+)$", answer):  # test letter-number format, e.g. "C4"
                reward = 0.05
            else:
                reward = 0.01
        return reward


class TsumegoCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(TsumegoCurriculum.__name__, TsumegoConfig)
        self._define_attributes(
            RangeAttributeDefinition(
                name="board_size",
                levels=[5, 10, 15, 19],
                lower_field_name="min_board_size",
                upper_field_name="max_board_size",
                description="The size of the board",
            ),
            ScalarAttributeDefinition(
                name="max_stones",
                field_name="max_stones",
                levels=[5, 10, 13, 15],
                description="The maximum number of stones on the board",
            ),
        )


# Register the dataset
register_dataset(DATASET_NAME, TsumegoDataset, TsumegoConfig, TsumegoCurriculum)



"""
Game tasks for training reasoning capabilities:
- Board games
- Puzzle games
- Strategy games
- Simulation games
"""

from .boxnet import BoxnetConfig, BoxnetCurriculum, BoxnetDataset
from .countdown import CountdownConfig, CountdownCurriculum, CountdownDataset
from .emoji_mystery import EmojiMysteryConfig, EmojiMysteryCurriculum, EmojiMysteryDataset
from .futoshiki import FutoshikiConfig, FutoshikiCurriculum, FutoshikiDataset
from .kakurasu import KakurasuConfig, KakurasuCurriculum, KakurasuDataset
from .knight_swap import KnightSwapConfig, KnightSwapCurriculum, KnightSwapDataset
from .mahjong import MahjongPuzzleConfig, MahjongPuzzleCurriculum, MahjongPuzzleDataset
from .maze import MazeConfig, MazeCurriculum, MazeDataset
from .mini_sudoku import MiniSudokuConfig, MiniSudokuCurriculum, MiniSudokuDataset
from .n_queens import NQueensConfig, NQueensCurriculum, NQueensDataset
from .puzzle24 import Puzzle24Config, Puzzle24Curriculum, Puzzle24Dataset
from .rush_hour import RushHourConfig, RushHourCurriculum, RushHourDataset
from .sokoban import SokobanConfig, SokobanCurriculum, SokobanDataset
from .sudoku import SudokuConfig, SudokuCurriculum, SudokuDataset
from .survo import SurvoConfig, SurvoCurriculum, SurvoDataset
from .tower_of_hanoi import HanoiConfig, HanoiCurriculum, HanoiDataset
from .tsumego import TsumegoConfig, TsumegoCurriculum, TsumegoDataset

__all__ = [
    "BoxnetConfig",
    "BoxnetDataset",
    "BoxnetCurriculum",
    "CountdownConfig",
    "CountdownDataset",
    "CountdownCurriculum",
    "EmojiMysteryConfig",
    "EmojiMysteryCurriculum",
    "EmojiMysteryDataset",
    "FutoshikiConfig",
    "FutoshikiCurriculum",
    "FutoshikiDataset",
    "KakurasuConfig",
    "KakurasuCurriculum",
    "KakurasuDataset",
    "MiniSudokuConfig",
    "MiniSudokuDataset",
    "MiniSudokuCurriculum",
    "Puzzle24Config",
    "Puzzle24Dataset",
    "Puzzle24Curriculum",
    "SokobanConfig",
    "SokobanCurriculum",
    "SokobanDataset",
    "SudokuConfig",
    "SudokuCurriculum",
    "SudokuDataset",
    "SurvoConfig",
    "SurvoCurriculum",
    "SurvoDataset",
    "RushHourConfig",
    "RushHourCurriculum",
    "RushHourDataset",
    "MazeConfig",
    "MazeDataset",
    "MazeCurriculum",
    "HanoiConfig",
    "HanoiDataset",
    "HanoiCurriculum",
    "NQueensDataset",
    "NQueensConfig",
    "NQueensCurriculum",
    "TsumegoConfig",
    "TsumegoCurriculum",
    "TsumegoDataset",
    "KnightSwapConfig",
    "KnightSwapDataset",
    "KnightSwapCurriculum",
    "MahjongPuzzleConfig",
    "MahjongPuzzleDataset",
    "MahjongPuzzleCurriculum",
]



from dataclasses import dataclass, field
from random import Random
from typing import Any, Callable, Optional

from ..coaching import BaseCurriculum, ScalarAttributeDefinition
from ..factory import ProceduralDataset, register_dataset
from .board_format import ARC_PROMPT_TEMPLATE, BoardFormattingOptions, format_board, format_board_pair, parse_board

RNG_DIFFICULTY_LEVELS = [0.0, 0.025, 0.05, 0.075, 0.1, 0.125, 0.15, 0.2]
RNG_DIFFICULTY_RANGES = [
    (RNG_DIFFICULTY_LEVELS[i], RNG_DIFFICULTY_LEVELS[i + 1]) for i in range(len(RNG_DIFFICULTY_LEVELS) - 1)
]

PSO_DIFFICULTY_LEVELS = [0.05, 0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 1]
PSO_DIFFICULTY_RANGES = [
    (PSO_DIFFICULTY_LEVELS[i], PSO_DIFFICULTY_LEVELS[i + 1]) for i in range(len(PSO_DIFFICULTY_LEVELS) - 1)
]

DATASET_NAME = "rearc"


@dataclass
class ReArcConfig:
    min_examples: int = 3  # minimum number of board pairs shown
    max_examples: int = 5  # maximum number of board pairs shown
    diff_lb: int = 0
    diff_ub: int = 0.2
    board_format_opts: BoardFormattingOptions = field(default_factory=lambda: BoardFormattingOptions())
    seed: Optional[int] = None
    size: int = 500
    rng_difficulty_ranges: list[tuple[float, float]] = field(default_factory=lambda: RNG_DIFFICULTY_RANGES)
    rng_difficulty_weights: list[float] = field(
        default_factory=lambda: [1 / len(RNG_DIFFICULTY_RANGES)] * len(RNG_DIFFICULTY_RANGES)
    )
    pso_difficulty_ranges: list[tuple[float, float]] = field(default_factory=lambda: PSO_DIFFICULTY_RANGES)
    pso_difficulty_weights: list[float] = field(
        default_factory=lambda: [1 / len(PSO_DIFFICULTY_RANGES)] * len(PSO_DIFFICULTY_RANGES)
    )

    def validate(self):
        assert self.min_examples > 0, "min_examples must be positive"
        assert self.min_examples <= self.max_examples, "min_examples must be <= max_examples"
        assert self.diff_lb <= self.diff_ub, "diff_lb must be <= diff_ub."
        assert self.size > 0, "Size of dataset must be positive."
        assert len(self.rng_difficulty_ranges) == len(
            self.rng_difficulty_weights
        ), "rng_difficulty_ranges and rng_difficulty_weights must have the same length."
        assert len(self.pso_difficulty_ranges) == len(
            self.pso_difficulty_weights
        ), "pso_difficulty_ranges and pso_difficulty_weights must have the same length."


class ReArcDataset(ProceduralDataset):
    def __init__(self, config: ReArcConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)
        self.board_format_opts = config.board_format_opts
        self._prompt_templates = ARC_PROMPT_TEMPLATE
        self.diff_lb = config.diff_lb
        self.diff_ub = config.diff_ub

        # lazy import of re-arc dsl & generators
        from .rearc_utils import generators
        from .rearc_utils.utils import get_generators, get_pso_difficulty

        self._generators = get_generators(generators)
        self.get_pso_difficulty = get_pso_difficulty

    @staticmethod
    def get_rng_difficulty(rng: Random) -> float:
        if not hasattr(rng, "difficulty_samples"):
            return 0.0
        samples = rng.difficulty_samples
        avg = sum(samples) / len(samples) if samples else 0.0
        rng.difficulty_samples = []
        return avg

    def __len__(self) -> int:
        return self.size

    def format_rearc_input(self, rng: Random, task: dict, generator: Callable) -> str:
        """
        Format a ReArc task input with multiple examples and test input.
        """

        num_examples = rng.randint(self.config.min_examples, self.config.max_examples)
        examples = [
            format_board_pair(
                i + 1, generator(rng, self.diff_lb, self.diff_ub), formatting_options=self.config.board_format_opts
            )
            for i in range(num_examples)
        ]
        examples = "".join(examples)
        input_grid = format_board(task["input"], self.board_format_opts)

        return self._prompt_templates.format(examples=examples, input_grid=input_grid)

    def __getitem__(self, idx: int) -> dict:
        """
        Generate a single ReArc task
        """
        rng = Random(self.seed + idx)

        pso_difficulty_range = rng.choices(
            self.config.pso_difficulty_ranges, weights=self.config.pso_difficulty_weights, k=1
        )[0]

        while True:
            task_id = rng.choice(list(self._generators.keys()))
            generator = self._generators[task_id]
            difficulty_range = rng.choices(
                self.config.rng_difficulty_ranges, weights=self.config.rng_difficulty_weights, k=1
            )[0]
            task = generator(rng, difficulty_range[0], difficulty_range[1])
            pso_difficulty = self.get_pso_difficulty(task)
            if (pso_difficulty_range[0] <= pso_difficulty) and (pso_difficulty <= pso_difficulty_range[1]):
                break

        rng_difficulty = self.get_rng_difficulty(rng)
        input_prompt = self.format_rearc_input(rng, task, generator)
        answer = format_board(task["output"], self.board_format_opts)

        return {
            "question": input_prompt,
            "answer": answer,
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "input": task["input"],
                "output": task["output"],
                "task_id": task_id,
                "rng": rng_difficulty,
                "pso": pso_difficulty,
                "difficulty": {
                    "rng_difficulty_weights": self.config.rng_difficulty_weights,
                    "pso_difficulty_weights": self.config.pso_difficulty_weights,
                },
            },
        }

    def score_answer(self, answer: str, entry: dict[str, Any]) -> float:
        reward = 0.0
        metadata = entry["metadata"]
        if answer is not None:
            try:
                answer_board = parse_board(answer, self.board_format_opts)
                if answer_board == metadata["output"]:
                    reward = 1.0
                else:
                    reward = 0.05
            except:
                reward = 0.0
        return reward


class ReArcCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(ReArcCurriculum.__name__, ReArcConfig)
        self._define_attributes(
            ScalarAttributeDefinition(
                name="pso_difficulty_weights",
                field_name="pso_difficulty_weights",
                description="The range of PSO difficulty for the Arc problem",
                levels=[
                    [1, 0, 0, 0, 0, 0, 0],  # only sample/generate the easiest tasks wrs PSO difficulty
                    [0, 1, 0, 0, 0, 0, 0],
                    [0, 0, 1, 0, 0, 0, 0],
                    [0, 0, 0, 1, 0, 0, 0],
                    [0, 0, 0, 0, 1, 0, 0],
                    [0, 0, 0, 0, 0, 1, 0],
                    [0, 0, 0, 0, 0, 0, 1],
                ],  # only sample/generate the hardest tasks PSO difficulty
            ),
            ScalarAttributeDefinition(
                name="rng_difficulty_weights",
                field_name="rng_difficulty_weights",
                description="The range of RNG difficulty for the Arc problem",
                levels=[
                    [1, 0, 0, 0, 0, 0, 0],  # only sample/generate the easiest tasks wrs RNG difficulty
                    [0, 1, 0, 0, 0, 0, 0],
                    [0, 0, 1, 0, 0, 0, 0],
                    [0, 0, 0, 1, 0, 0, 0],
                    [0, 0, 0, 0, 1, 0, 0],
                    [0, 0, 0, 0, 0, 1, 0],
                    [0, 0, 0, 0, 0, 0, 1],
                ],  # only sample/generate the hardest tasks wrs RNG difficulty
            ),
        )


register_dataset(DATASET_NAME, ReArcDataset, ReArcConfig, ReArcCurriculum)
